from integrations.postgres import PostgresConnection


async def execute_migration() -> None:
    await PostgresConnection.execute(
        """\
        CREATE TABLE IF NOT EXISTS
            todos (
                id UUID NOT NULL DEFAULT gen_random_uuid(),
                modified TIMESTAMP WITH TIME ZONE NOT NULL DEFAULT now(),
                description TEXT NOT NULL,
                completed BOOL NOT NULL DEFAULT false,
                PRIMARY KEY(id)
            )
        ;\
        """
    )



from integrations.postgres.client import PostgresConnectionPool
from integrations.postgres.state import Postgres, PostgresConnection
from integrations.postgres.types import PostgresException, PostgresRow, PostgresValue

__all__ = [
    "Postgres",
    "PostgresConnection",
    "PostgresConnectionPool",
    "PostgresException",
    "PostgresRow",
    "PostgresValue",
]



from contextvars import ContextVar, Token
from logging import DEBUG, ERROR, INFO, WARNING, Logger, getLogger
from time import monotonic
from types import TracebackType
from typing import Any, Self, final

from haiway.context.identifier import ScopeIdentifier

__all__ = [
    "LoggerContext",
]


@final
class LoggerContext:
    _context = ContextVar[Self]("LoggerContext")

    @classmethod
    def scope(
        cls,
        scope: ScopeIdentifier,
        /,
        *,
        logger: Logger | None,
    ) -> Self:
        current: Self
        try:  # check for current scope
            current = cls._context.get()

        except LookupError:
            # create root scope when missing
            return cls(
                scope=scope,
                logger=logger,
            )

        # create nested scope otherwise
        return cls(
            scope=scope,
            logger=logger or current._logger,
        )

    @classmethod
    def log_error(
        cls,
        message: str,
        /,
        *args: Any,
        exception: BaseException | None = None,
    ) -> None:
        try:
            cls._context.get().log(
                ERROR,
                message,
                *args,
                exception=exception,
            )

        except LookupError:
            getLogger().log(
                ERROR,
                message,
                *args,
                exc_info=exception,
            )

    @classmethod
    def log_warning(
        cls,
        message: str,
        /,
        *args: Any,
        exception: Exception | None = None,
    ) -> None:
        try:
            cls._context.get().log(
                WARNING,
                message,
                *args,
                exception=exception,
            )

        except LookupError:
            getLogger().log(
                WARNING,
                message,
                *args,
                exc_info=exception,
            )

    @classmethod
    def log_info(
        cls,
        message: str,
        /,
        *args: Any,
    ) -> None:
        try:
            cls._context.get().log(
                INFO,
                message,
                *args,
            )

        except LookupError:
            getLogger().log(
                INFO,
                message,
                *args,
            )

    @classmethod
    def log_debug(
        cls,
        message: str,
        /,
        *args: Any,
        exception: Exception | None = None,
    ) -> None:
        try:
            cls._context.get().log(
                DEBUG,
                message,
                *args,
                exception=exception,
            )

        except LookupError:
            getLogger().log(
                DEBUG,
                message,
                *args,
                exc_info=exception,
            )

    __slots__ = (
        "_entered",
        "_logger",
        "_prefix",
        "_token",
    )

    def __init__(
        self,
        scope: ScopeIdentifier,
        logger: Logger | None,
    ) -> None:
        self._prefix: str
        object.__setattr__(
            self,
            "_prefix",
            scope.unique_name,
        )
        self._logger: Logger
        object.__setattr__(
            self,
            "_logger",
            logger or getLogger(name=scope.label),
        )
        self._token: Token[LoggerContext] | None
        object.__setattr__(
            self,
            "_token",
            None,
        )
        self._entered: float | None
        object.__setattr__(
            self,
            "_entered",
            None,
        )

    def __setattr__(
        self,
        name: str,
        value: Any,
    ) -> Any:
        raise AttributeError(
            f"Can't modify immutable {self.__class__.__qualname__},"
            f" attribute - '{name}' cannot be modified"
        )

    def __delattr__(
        self,
        name: str,
    ) -> None:
        raise AttributeError(
            f"Can't modify immutable {self.__class__.__qualname__},"
            f" attribute - '{name}' cannot be deleted"
        )

    def log(
        self,
        level: int,
        message: str,
        /,
        *args: Any,
        exception: BaseException | None = None,
    ) -> None:
        self._logger.log(
            level,
            f"{self._prefix} {message}",
            *args,
            exc_info=exception,
        )

    def __enter__(self) -> None:
        assert self._token is None, "Context reentrance is not allowed"  # nosec: B101
        assert self._entered is None, "Context reentrance is not allowed"  # nosec: B101
        object.__setattr__(
            self,
            "_token",
            LoggerContext._context.set(self),
        )
        object.__setattr__(
            self,
            "_entered",
            monotonic(),
        )
        self.log(DEBUG, "Entering context...")

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        assert (  # nosec: B101
            self._token is not None and self._entered is not None
        ), "Unbalanced context enter/exit"
        LoggerContext._context.reset(self._token)
        object.__setattr__(
            self,
            "_token",
            None,
        )
        self.log(DEBUG, f"...exiting context after {monotonic() - self._entered:.2f}s")
        object.__setattr__(
            self,
            "_entered",
            None,
        )



from collections.abc import Callable, Mapping, MutableMapping, Sequence, Set
from datetime import date, datetime, time, timedelta, timezone
from enum import Enum
from pathlib import Path
from re import Pattern
from types import EllipsisType, NoneType, UnionType
from typing import Any, Literal, Protocol, Self, Union, final, is_typeddict
from uuid import UUID

from haiway.state.attributes import AttributeAnnotation
from haiway.types import MISSING, Missing

__all__ = [
    "AttributeValidation",
    "AttributeValidationError",
    "AttributeValidator",
]


class AttributeValidation[Type](Protocol):
    def __call__(
        self,
        value: Any,
        /,
    ) -> Type: ...


class AttributeValidationError(Exception):
    pass


@final
class AttributeValidator[Type]:
    @classmethod
    def of(
        cls,
        annotation: AttributeAnnotation,
        /,
        *,
        recursion_guard: MutableMapping[str, AttributeValidation[Any]],
    ) -> AttributeValidation[Any]:
        if isinstance(annotation.origin, NotImplementedError | RuntimeError):
            raise annotation.origin  # raise an error if origin was not properly resolved

        if recursive := recursion_guard.get(str(annotation)):
            return recursive

        validator: Self = cls(
            annotation,
            validation=MISSING,
        )
        recursion_guard[str(annotation)] = validator

        if common := VALIDATORS.get(annotation.origin):
            object.__setattr__(
                validator,
                "validation",
                common(annotation, recursion_guard),
            )

        elif hasattr(annotation.origin, "__IMMUTABLE__"):
            object.__setattr__(
                validator,
                "validation",
                _prepare_validator_of_type(annotation, recursion_guard),
            )

        elif is_typeddict(annotation.origin):
            object.__setattr__(
                validator,
                "validation",
                _prepare_validator_of_typed_dict(annotation, recursion_guard),
            )

        elif issubclass(annotation.origin, Protocol):
            object.__setattr__(
                validator,
                "validation",
                _prepare_validator_of_type(annotation, recursion_guard),
            )

        elif issubclass(annotation.origin, Enum):
            object.__setattr__(
                validator,
                "validation",
                _prepare_validator_of_type(annotation, recursion_guard),
            )

        else:
            raise TypeError(f"Unsupported type annotation: {annotation}")

        return validator

    __slots__ = (
        "annotation",
        "validation",
    )

    def __init__(
        self,
        annotation: AttributeAnnotation,
        validation: AttributeValidation[Type] | Missing,
    ) -> None:
        self.annotation: AttributeAnnotation
        object.__setattr__(
            self,
            "annotation",
            annotation,
        )
        self.validation: AttributeValidation[Type] | Missing
        object.__setattr__(
            self,
            "validation",
            validation,
        )

    def __setattr__(
        self,
        name: str,
        value: Any,
    ) -> Any:
        raise AttributeError(
            f"Can't modify immutable {self.__class__.__qualname__},"
            f" attribute - '{name}' cannot be modified"
        )

    def __delattr__(
        self,
        name: str,
    ) -> None:
        raise AttributeError(
            f"Can't modify immutable {self.__class__.__qualname__},"
            f" attribute - '{name}' cannot be deleted"
        )

    def __call__(
        self,
        value: Any,
        /,
    ) -> Any:
        assert self.validation is not MISSING  # nosec: B101
        return self.validation(value)  # pyright: ignore[reportCallIssue, reportUnknownVariableType]

    def __str__(self) -> str:
        return f"Validator[{self.annotation}]"

    def __repr__(self) -> str:
        return f"Validator[{self.annotation}]"


def _prepare_validator_of_any(
    annotation: AttributeAnnotation,
    /,
    recursion_guard: MutableMapping[str, AttributeValidation[Any]],
) -> AttributeValidation[Any]:
    def validator(
        value: Any,
        /,
    ) -> Any:
        return value  # any is always valid

    return validator


def _prepare_validator_of_none(
    annotation: AttributeAnnotation,
    /,
    recursion_guard: MutableMapping[str, AttributeValidation[Any]],
) -> AttributeValidation[Any]:
    def validator(
        value: Any,
        /,
    ) -> Any:
        if value is None:
            return value

        else:
            raise TypeError(f"'{value}' is not matching expected type of 'None'")

    return validator


def _prepare_validator_of_missing(
    annotation: AttributeAnnotation,
    /,
    recursion_guard: MutableMapping[str, AttributeValidation[Any]],
) -> AttributeValidation[Any]:
    def validator(
        value: Any,
        /,
    ) -> Any:
        if value is MISSING:
            return value

        else:
            raise TypeError(f"'{value}' is not matching expected type of 'Missing'")

    return validator


def _prepare_validator_of_literal(
    annotation: AttributeAnnotation,
    /,
    recursion_guard: MutableMapping[str, AttributeValidation[Any]],
) -> AttributeValidation[Any]:
    elements: Sequence[Any] = annotation.arguments
    formatted_type: str = str(annotation)

    def validator(
        value: Any,
        /,
    ) -> Any:
        if value in elements:
            return value

        else:
            raise ValueError(f"'{value}' is not matching expected values of '{formatted_type}'")

    return validator


def _prepare_validator_of_type(
    annotation: AttributeAnnotation,
    /,
    recursion_guard: MutableMapping[str, AttributeValidation[Any]],
) -> AttributeValidation[Any]:
    validated_type: type[Any] = annotation.origin
    formatted_type: str = str(annotation)

    def validator(
        value: Any,
        /,
    ) -> Any:
        match value:
            case value if isinstance(value, validated_type):
                return value

            case _:
                raise TypeError(f"'{value}' is not matching expected type of '{formatted_type}'")

    return validator


def _prepare_validator_of_set(
    annotation: AttributeAnnotation,
    /,
    recursion_guard: MutableMapping[str, AttributeValidation[Any]],
) -> AttributeValidation[Any]:
    element_validator: AttributeValidation[Any] = AttributeValidator.of(
        annotation.arguments[0],
        recursion_guard=recursion_guard,
    )
    formatted_type: str = str(annotation)

    def validator(
        value: Any,
        /,
    ) -> Any:
        if isinstance(value, set):
            return frozenset(element_validator(element) for element in value)  # pyright: ignore[reportUnknownVariableType]

        else:
            raise TypeError(f"'{value}' is not matching expected type of '{formatted_type}'")

    return validator


def _prepare_validator_of_sequence(
    annotation: AttributeAnnotation,
    /,
    recursion_guard: MutableMapping[str, AttributeValidation[Any]],
) -> AttributeValidation[Any]:
    element_validator: AttributeValidation[Any] = AttributeValidator.of(
        annotation.arguments[0],
        recursion_guard=recursion_guard,
    )
    formatted_type: str = str(annotation)

    def validator(
        value: Any,
        /,
    ) -> Any:
        match value:
            case [*elements]:
                return tuple(element_validator(element) for element in elements)

            case _:
                raise TypeError(f"'{value}' is not matching expected type of '{formatted_type}'")

    return validator


def _prepare_validator_of_mapping(
    annotation: AttributeAnnotation,
    /,
    recursion_guard: MutableMapping[str, AttributeValidation[Any]],
) -> AttributeValidation[Any]:
    key_validator: AttributeValidation[Any] = AttributeValidator.of(
        annotation.arguments[0],
        recursion_guard=recursion_guard,
    )
    value_validator: AttributeValidation[Any] = AttributeValidator.of(
        annotation.arguments[1],
        recursion_guard=recursion_guard,
    )
    formatted_type: str = str(annotation)

    def validator(
        value: Any,
        /,
    ) -> Any:
        match value:
            case {**elements}:
                # TODO: make sure dict is not mutable with MappingProxyType?
                return {
                    key_validator(key): value_validator(element)
                    for key, element in elements.items()
                }

            case _:
                raise TypeError(f"'{value}' is not matching expected type of '{formatted_type}'")

    return validator


def _prepare_validator_of_tuple(
    annotation: AttributeAnnotation,
    /,
    recursion_guard: MutableMapping[str, AttributeValidation[Any]],
) -> AttributeValidation[Any]:
    if (
        annotation.arguments[-1].origin == Ellipsis
        or annotation.arguments[-1].origin == EllipsisType
    ):
        element_validator: AttributeValidation[Any] = AttributeValidator.of(
            annotation.arguments[0],
            recursion_guard=recursion_guard,
        )
        formatted_type: str = str(annotation)

        def validator(
            value: Any,
            /,
        ) -> Any:
            match value:
                case [*elements]:
                    return tuple(element_validator(element) for element in elements)

                case _:
                    raise TypeError(
                        f"'{value}' is not matching expected type of '{formatted_type}'"
                    )

        return validator

    else:
        element_validators: list[AttributeValidation[Any]] = [
            AttributeValidator.of(alternative, recursion_guard=recursion_guard)
            for alternative in annotation.arguments
        ]
        elements_count: int = len(element_validators)
        formatted_type: str = str(annotation)

        def validator(
            value: Any,
            /,
        ) -> Any:
            match value:
                case [*elements]:
                    if len(elements) != elements_count:
                        raise ValueError(
                            f"'{value}' is not matching expected type of '{formatted_type}'"
                        )

                    return tuple(
                        element_validators[idx](element) for idx, element in enumerate(elements)
                    )

                case _:
                    raise TypeError(
                        f"'{value}' is not matching expected type of '{formatted_type}'"
                    )

        return validator


def _prepare_validator_of_union(
    annotation: AttributeAnnotation,
    /,
    recursion_guard: MutableMapping[str, AttributeValidation[Any]],
) -> AttributeValidation[Any]:
    validators: list[AttributeValidation[Any]] = [
        AttributeValidator.of(alternative, recursion_guard=recursion_guard)
        for alternative in annotation.arguments
    ]
    formatted_type: str = str(annotation)

    def validator(
        value: Any,
        /,
    ) -> Any:
        errors: list[Exception] = []
        for validator in validators:
            try:
                return validator(value)

            except Exception as exc:
                errors.append(exc)

        raise ExceptionGroup(
            f"'{value}' is not matching expected type of '{formatted_type}'",
            errors,
        )

    return validator


def _prepare_validator_of_callable(
    annotation: AttributeAnnotation,
    /,
    recursion_guard: MutableMapping[str, AttributeValidation[Any]],
) -> AttributeValidation[Any]:
    formatted_type: str = str(annotation)

    def validator(
        value: Any,
        /,
    ) -> Any:
        if callable(value):
            # TODO: we could verify callable signature here
            return value

        else:
            raise TypeError(f"'{value}' is not matching expected type of '{formatted_type}'")

    return validator


def _prepare_validator_of_typed_dict(
    annotation: AttributeAnnotation,
    /,
    recursion_guard: MutableMapping[str, AttributeValidation[Any]],
) -> AttributeValidation[Any]:
    def key_validator(
        value: Any,
        /,
    ) -> Any:
        match value:
            case value if isinstance(value, str):
                return value

            case _:
                raise TypeError(f"'{value}' is not matching expected type of 'str'")

    formatted_type: str = str(annotation)
    values_validators: dict[str, AttributeValidation[Any]] = {
        key: AttributeValidator.of(element, recursion_guard=recursion_guard)
        for key, element in annotation.extra["attributes"].items()
    }
    required_values: Set[str] = annotation.extra["required"]

    def validator(
        value: Any,
        /,
    ) -> Any:
        match value:
            case {**elements}:
                validated: MutableMapping[str, Any] = {}
                for key, validator in values_validators.items():
                    element: Any = elements.get(key, MISSING)
                    if element is MISSING and key not in required_values:
                        continue  # skip missing and not required

                    validated[key_validator(key)] = validator(element)

                # TODO: make sure dict is not mutable with MappingProxyType?
                return validated

            case _:
                raise TypeError(f"'{value}' is not matching expected type of '{formatted_type}'")

    return validator


VALIDATORS: Mapping[
    Any,
    Callable[
        [
            AttributeAnnotation,
            MutableMapping[str, AttributeValidation[Any]],
        ],
        AttributeValidation[Any],
    ],
] = {
    Any: _prepare_validator_of_any,
    NoneType: _prepare_validator_of_none,
    Missing: _prepare_validator_of_missing,
    type: _prepare_validator_of_type,
    bool: _prepare_validator_of_type,
    int: _prepare_validator_of_type,
    float: _prepare_validator_of_type,
    complex: _prepare_validator_of_type,
    bytes: _prepare_validator_of_type,
    str: _prepare_validator_of_type,
    tuple: _prepare_validator_of_tuple,
    frozenset: _prepare_validator_of_set,
    Set: _prepare_validator_of_set,
    Sequence: _prepare_validator_of_sequence,
    Mapping: _prepare_validator_of_mapping,
    Literal: _prepare_validator_of_literal,
    range: _prepare_validator_of_type,
    UUID: _prepare_validator_of_type,
    date: _prepare_validator_of_type,
    datetime: _prepare_validator_of_type,
    time: _prepare_validator_of_type,
    timedelta: _prepare_validator_of_type,
    timezone: _prepare_validator_of_type,
    Path: _prepare_validator_of_type,
    Pattern: _prepare_validator_of_type,
    Union: _prepare_validator_of_union,
    UnionType: _prepare_validator_of_union,
    Callable: _prepare_validator_of_callable,
}



from haiway.utils.always import always, async_always
from haiway.utils.collections import as_dict, as_list, as_set, as_tuple
from haiway.utils.env import (
    getenv_base64,
    getenv_bool,
    getenv_float,
    getenv_int,
    getenv_str,
    load_env,
)
from haiway.utils.freezing import freeze
from haiway.utils.logs import setup_logging
from haiway.utils.mimic import mimic_function
from haiway.utils.noop import async_noop, noop
from haiway.utils.queue import AsyncQueue
from haiway.utils.stream import AsyncStream

__all__ = [
    "AsyncQueue",
    "AsyncStream",
    "always",
    "as_dict",
    "as_list",
    "as_set",
    "as_tuple",
    "async_always",
    "async_noop",
    "freeze",
    "getenv_base64",
    "getenv_bool",
    "getenv_float",
    "getenv_int",
    "getenv_str",
    "load_env",
    "mimic_function",
    "noop",
    "setup_logging",
]



from collections.abc import Callable, Collection, Iterable
from typing import Any, Literal, Self, cast, final

from haiway.state.path import AttributePath

__all__ = [
    "AttributeRequirement",
]


@final
class AttributeRequirement[Root]:
    @classmethod
    def equal[Parameter](
        cls,
        value: Parameter,
        /,
        path: AttributePath[Root, Parameter] | Parameter,
    ) -> Self:
        assert isinstance(  # nosec: B101
            path, AttributePath
        ), "Prepare attribute path by using Self._.path.to.property or explicitly"

        def check_equal(root: Root) -> None:
            checked: Any = cast(AttributePath[Root, Parameter], path)(root)
            if checked != value:
                raise ValueError(f"{checked} is not equal {value} for '{path.__repr__()}'")

        return cls(
            path,
            "equal",
            value,
            check=check_equal,
        )

    @classmethod
    def not_equal[Parameter](
        cls,
        value: Parameter,
        /,
        path: AttributePath[Root, Parameter] | Parameter,
    ) -> Self:
        assert isinstance(  # nosec: B101
            path, AttributePath
        ), "Prepare attribute path by using Self._.path.to.property or explicitly"

        def check_not_equal(root: Root) -> None:
            checked: Any = cast(AttributePath[Root, Parameter], path)(root)
            if checked == value:
                raise ValueError(f"{checked} is equal {value} for '{path.__repr__()}'")

        return cls(
            path,
            "not_equal",
            value,
            check=check_not_equal,
        )

    @classmethod
    def contains[Parameter](
        cls,
        value: Parameter,
        /,
        path: AttributePath[
            Root,
            Collection[Parameter] | tuple[Parameter, ...] | list[Parameter] | set[Parameter],
        ]
        | Collection[Parameter]
        | tuple[Parameter, ...]
        | list[Parameter]
        | set[Parameter],
    ) -> Self:
        assert isinstance(  # nosec: B101
            path, AttributePath
        ), "Prepare attribute path by using Self._.path.to.property or explicitly"

        def check_contains(root: Root) -> None:
            checked: Any = cast(AttributePath[Root, Parameter], path)(root)
            if value not in checked:
                raise ValueError(f"{checked} does not contain {value} for '{path.__repr__()}'")

        return cls(
            path,
            "contains",
            value,
            check=check_contains,
        )

    @classmethod
    def contains_any[Parameter](
        cls,
        value: Collection[Parameter],
        /,
        path: AttributePath[
            Root,
            Collection[Parameter] | tuple[Parameter, ...] | list[Parameter] | set[Parameter],
        ]
        | Collection[Parameter]
        | tuple[Parameter, ...]
        | list[Parameter]
        | set[Parameter],
    ) -> Self:
        assert isinstance(  # nosec: B101
            path, AttributePath
        ), "Prepare attribute path by using Self._.path.to.property or explicitly"

        def check_contains_any(root: Root) -> None:
            checked: Any = cast(AttributePath[Root, Parameter], path)(root)
            if any(element in checked for element in value):
                raise ValueError(
                    f"{checked} does not contain any of {value} for '{path.__repr__()}'"
                )

        return cls(
            path,
            "contains_any",
            value,
            check=check_contains_any,
        )

    @classmethod
    def contained_in[Parameter](
        cls,
        value: Collection[Parameter],
        /,
        path: AttributePath[Root, Parameter] | Parameter,
    ) -> Self:
        assert isinstance(  # nosec: B101
            path, AttributePath
        ), "Prepare attribute path by using Self._.path.to.property or explicitly"

        def check_contained_in(root: Root) -> None:
            checked: Any = cast(AttributePath[Root, Parameter], path)(root)
            if checked not in value:
                raise ValueError(f"{value} does not contain {checked} for '{path.__repr__()}'")

        return cls(
            value,
            "contained_in",
            path,
            check=check_contained_in,
        )

    __slots__ = (
        "_check",
        "lhs",
        "operator",
        "rhs",
    )

    def __init__(
        self,
        lhs: Any,
        operator: Literal[
            "equal",
            "not_equal",
            "contains",
            "contains_any",
            "contained_in",
            "and",
            "or",
        ],
        rhs: Any,
        check: Callable[[Root], None],
    ) -> None:
        self.lhs: Any
        object.__setattr__(
            self,
            "lhs",
            lhs,
        )
        self.operator: Literal[
            "equal",
            "not_equal",
            "contains",
            "contains_any",
            "contained_in",
            "and",
            "or",
        ]
        object.__setattr__(
            self,
            "operator",
            operator,
        )
        self.rhs: Any
        object.__setattr__(
            self,
            "rhs",
            rhs,
        )
        self._check: Callable[[Root], None]
        object.__setattr__(
            self,
            "_check",
            check,
        )

    def __and__(
        self,
        other: Self,
    ) -> Self:
        def check_and(root: Root) -> None:
            self.check(root)
            other.check(root)

        return self.__class__(
            self,
            "and",
            other,
            check=check_and,
        )

    def __or__(
        self,
        other: Self,
    ) -> Self:
        def check_or(root: Root) -> None:
            try:
                self.check(root)
            except ValueError:
                other.check(root)

        return self.__class__(
            self,
            "or",
            other,
            check=check_or,
        )

    def check(
        self,
        root: Root,
        /,
        *,
        raise_exception: bool = True,
    ) -> bool:
        try:
            self._check(root)
            return True

        except Exception as exc:
            if raise_exception:
                raise exc

            else:
                return False

    def filter(
        self,
        values: Iterable[Root],
    ) -> list[Root]:
        return [value for value in values if self.check(value, raise_exception=False)]

    def __setattr__(
        self,
        name: str,
        value: Any,
    ) -> Any:
        raise AttributeError(
            f"Can't modify immutable {self.__class__.__qualname__},"
            f" attribute - '{name}' cannot be modified"
        )

    def __delattr__(
        self,
        name: str,
    ) -> None:
        raise AttributeError(
            f"Can't modify immutable {self.__class__.__qualname__},"
            f" attribute - '{name}' cannot be deleted"
        )



from fastapi import APIRouter
from starlette.responses import Response

__all__ = [
    "router",
]

router = APIRouter()


@router.get(
    path="/health",
    description="Server health check.",
    status_code=204,
    responses={
        204: {"description": "Server up and running!"},
        500: {"description": "Internal server error"},
    },
)
async def health() -> Response:
    return Response(status_code=204)



from collections.abc import Sequence
from datetime import datetime
from typing import Protocol, overload, runtime_checkable
from uuid import UUID

from haiway import State

__all__ = [
    "UserTask",
    "UserTaskCreation",
    "UserTaskDeletion",
    "UserTaskFetching",
    "UserTaskUpdating",
]


class UserTask(State):
    identifier: UUID
    modified: datetime
    description: str
    completed: bool


@runtime_checkable
class UserTaskCreation(Protocol):
    async def __call__(
        self,
        *,
        description: str,
    ) -> UserTask: ...


@runtime_checkable
class UserTaskUpdating(Protocol):
    async def __call__(
        self,
        *,
        task: UserTask,
    ) -> None: ...


@runtime_checkable
class UserTaskFetching(Protocol):
    @overload
    async def __call__(
        self,
        *,
        identifier: None = None,
    ) -> Sequence[UserTask]: ...

    @overload
    async def __call__(
        self,
        *,
        identifier: UUID,
    ) -> UserTask: ...

    async def __call__(
        self,
        *,
        identifier: UUID | None = None,
    ) -> Sequence[UserTask] | UserTask: ...


@runtime_checkable
class UserTaskDeletion(Protocol):
    async def __call__(
        self,
        *,
        identifier: UUID,
    ) -> None: ...



from contextlib import asynccontextmanager
from logging import Logger, getLogger

from fastapi import FastAPI
from haiway import Disposables, setup_logging

from integrations.postgres import PostgresConnectionPool
from server.middlewares import ContextMiddleware
from server.routes import technical_router, todos_router

__all__ = [
    "app",
]


@asynccontextmanager
async def lifespan(app: FastAPI):
    setup_logging("server")

    logger: Logger = getLogger("server")
    if __debug__:
        logger.warning("Starting DEBUG server...")

    else:
        logger.info("Starting server...")

    disposables = Disposables(
        PostgresConnectionPool(),
    )
    async with disposables as state:
        app.extra["state"] = (*state,)

        logger.info("...server started...")
        yield  # suspend until server shutdown

    logger.info("...server shutdown!")


app: FastAPI = FastAPI(
    title="haiway-fastapi",
    description="Example project using haiway with fastapi",
    version="0.1.0",
    lifespan=lifespan,
    openapi_url="/openapi.json" if __debug__ else None,
    docs_url="/swagger" if __debug__ else None,
    redoc_url="/redoc" if __debug__ else None,
)

# middlewares
app.add_middleware(ContextMiddleware)

# routes
app.include_router(technical_router)
app.include_router(todos_router)



from collections.abc import Sequence
from typing import Any

from haiway import State, ctx

from integrations.postgres.types import (
    PostgresConnectionContext,
    PostgresConnectionPreparing,
    PostgresRow,
    PostgresStatementExecuting,
    PostgresTransactionContext,
    PostgresTransactionPreparing,
)

__all__ = [
    "Postgres",
    "PostgresConnection",
]


class PostgresConnection(State):
    execute_statement: PostgresStatementExecuting
    prepare_transaction: PostgresTransactionPreparing

    @classmethod
    async def fetch_one(
        cls,
        statement: str,
        /,
        *args: Any,
    ) -> PostgresRow | None:
        return next(
            (
                result
                for result in await ctx.state(cls).execute_statement(
                    statement,
                    *args,
                )
            ),
            None,
        )

    @classmethod
    async def fetch(
        cls,
        statement: str,
        /,
        *args: Any,
    ) -> Sequence[PostgresRow]:
        return await ctx.state(cls).execute_statement(
            statement,
            *args,
        )

    @classmethod
    async def execute(
        cls,
        statement: str,
        /,
        *args: Any,
    ) -> None:
        await ctx.state(cls).execute_statement(
            statement,
            *args,
        )

    @classmethod
    def transaction(cls) -> PostgresTransactionContext:
        return ctx.state(cls).prepare_transaction()


class Postgres(State):
    @classmethod
    def connection(cls) -> PostgresConnectionContext:
        return ctx.state(cls).prepare_connection()

    prepare_connection: PostgresConnectionPreparing



from uuid import UUID

from haiway import State
from haiway.context import ctx

from features.todos.types import TodoCompletion
from features.todos.user_tasks import complete_todo_task

__all__ = [
    "Todos",
]


class Todos(State):
    @classmethod
    async def complete_todo(
        cls,
        *,
        identifier: UUID,
    ) -> None:
        await ctx.state(cls).complete(identifier=identifier)

    complete: TodoCompletion = complete_todo_task



from collections.abc import Sequence
from itertools import chain
from time import monotonic
from typing import Any, Self, cast, final, overload

from haiway.context import MetricsHandler, ScopeIdentifier, ctx
from haiway.state import State
from haiway.types import MISSING

__all_ = [
    "MetricsLogger",
    "MetricsHolder",
]


class MetricsScopeStore:
    __slots__ = (
        "allow_exit",
        "entered",
        "exited",
        "identifier",
        "metrics",
        "nested",
    )

    def __init__(
        self,
        identifier: ScopeIdentifier,
        /,
    ) -> None:
        self.identifier: ScopeIdentifier = identifier
        self.entered: float = monotonic()
        self.metrics: dict[type[State], State] = {}
        self.exited: float | None = None
        self.allow_exit: bool = False
        self.nested: list[MetricsScopeStore] = []

    @property
    def time(self) -> float:
        return (self.exited or monotonic()) - self.entered

    @property
    def finished(self) -> float:
        return self.exited is not None and all(nested.finished for nested in self.nested)

    @overload
    def merged[Metric: State](
        self,
    ) -> Sequence[State]: ...

    @overload
    def merged[Metric: State](
        self,
        metric: type[Metric],
    ) -> Metric | None: ...

    def merged[Metric: State](
        self,
        metric: type[Metric] | None = None,
    ) -> Sequence[State] | Metric | None:
        if metric is None:
            merged_metrics: dict[type[State], State] = dict(self.metrics)
            for nested in chain.from_iterable(nested.merged() for nested in self.nested):
                metric_type: type[State] = type(nested)
                current: State | None = merged_metrics.get(metric_type)

                if current is None:
                    merged_metrics[metric_type] = nested
                    continue  # keep going

                if hasattr(current, "__add__"):
                    merged_metrics[metric_type] = current.__add__(nested)  # pyright: ignore[reportUnknownMemberType, reportAttributeAccessIssue]
                    assert isinstance(merged_metrics[metric_type], State)  # nosec: B101
                    continue  # keep going

                break  # we have multiple value without a way to merge

            return tuple(merged_metrics.values())

        else:
            merged_metric: State | None = self.metrics.get(metric)
            for nested in self.nested:
                nested_metric: Metric | None = nested.merged(metric)
                if nested_metric is None:
                    continue  # skip missing

                if merged_metric is None:
                    merged_metric = nested_metric
                    continue  # keep going

                if hasattr(merged_metric, "__add__"):
                    merged_metric = merged_metric.__add__(nested_metric)  # pyright: ignore[reportUnknownMemberType, reportAttributeAccessIssue, reportUnknownVariableType]
                    assert isinstance(merged_metric, metric)  # nosec: B101
                    continue  # keep going

                break  # we have multiple value without a way to merge

            return cast(Metric | None, merged_metric)


@final
class MetricsHolder:
    @classmethod
    def handler(cls) -> MetricsHandler:
        store_handler: Self = cls()
        return MetricsHandler(
            record=store_handler.record,
            read=store_handler.read,
            enter_scope=store_handler.enter_scope,
            exit_scope=store_handler.exit_scope,
        )

    __slots__ = (
        "root_scope",
        "scopes",
    )

    def __init__(self) -> None:
        self.root_scope: ScopeIdentifier | None = None
        self.scopes: dict[str, MetricsScopeStore] = {}

    def record(
        self,
        scope: ScopeIdentifier,
        /,
        metric: State,
    ) -> None:
        assert self.root_scope is not None  # nosec: B101
        assert scope.scope_id in self.scopes  # nosec: B101

        metric_type: type[State] = type(metric)
        metrics: dict[type[State], State] = self.scopes[scope.scope_id].metrics
        if (current := metrics.get(metric_type)) and hasattr(current, "__add__"):
            metrics[type(metric)] = current.__add__(metric)  # pyright: ignore[reportUnknownMemberType, reportAttributeAccessIssue]

        metrics[type(metric)] = metric

    async def read[Metric: State](
        self,
        scope: ScopeIdentifier,
        /,
        *,
        metric: type[Metric],
        merged: bool,
    ) -> Metric | None:
        assert self.root_scope is not None  # nosec: B101
        assert scope.scope_id in self.scopes  # nosec: B101

        if merged:
            return self.scopes[scope.scope_id].merged(metric)

        else:
            return cast(Metric | None, self.scopes[scope.scope_id].metrics.get(metric))

    def enter_scope[Metric: State](
        self,
        scope: ScopeIdentifier,
        /,
    ) -> None:
        assert scope.scope_id not in self.scopes  # nosec: B101
        scope_metrics = MetricsScopeStore(scope)
        self.scopes[scope.scope_id] = scope_metrics

        if self.root_scope is None:
            self.root_scope = scope

        else:
            for key in self.scopes.keys():
                if key == scope.parent_id:
                    self.scopes[key].nested.append(scope_metrics)
                    return

            ctx.log_debug(
                "Attempting to enter nested scope metrics without entering its parent first"
            )

    def exit_scope[Metric: State](
        self,
        scope: ScopeIdentifier,
        /,
    ) -> None:
        assert self.root_scope is not None  # nosec: B101
        assert scope.scope_id in self.scopes  # nosec: B101

        self.scopes[scope.scope_id].allow_exit = True

        if not all(nested.exited for nested in self.scopes[scope.scope_id].nested):
            return  # not completed yet

        self.scopes[scope.scope_id].exited = monotonic()

        if scope != self.root_scope and self.scopes[scope.parent_id].allow_exit:
            self.exit_scope(self.scopes[scope.parent_id].identifier)


@final
class MetricsLogger:
    @classmethod
    def handler(
        cls,
        items_limit: int | None = None,
        redact_content: bool = False,
    ) -> MetricsHandler:
        logger_handler: Self = cls(
            items_limit=items_limit,
            redact_content=redact_content,
        )
        return MetricsHandler(
            record=logger_handler.record,
            read=logger_handler.read,
            enter_scope=logger_handler.enter_scope,
            exit_scope=logger_handler.exit_scope,
        )

    __slots__ = (
        "items_limit",
        "redact_content",
        "root_scope",
        "scopes",
    )

    def __init__(
        self,
        items_limit: int | None,
        redact_content: bool,
    ) -> None:
        self.root_scope: ScopeIdentifier | None = None
        self.scopes: dict[str, MetricsScopeStore] = {}
        self.items_limit: int | None = items_limit
        self.redact_content: bool = redact_content

    def record(
        self,
        scope: ScopeIdentifier,
        /,
        metric: State,
    ) -> None:
        assert self.root_scope is not None  # nosec: B101
        assert scope.scope_id in self.scopes  # nosec: B101

        metric_type: type[State] = type(metric)
        metrics: dict[type[State], State] = self.scopes[scope.scope_id].metrics
        if (current := metrics.get(metric_type)) and hasattr(current, "__add__"):
            metrics[type(metric)] = current.__add__(metric)  # pyright: ignore[reportUnknownMemberType, reportAttributeAccessIssue]

        metrics[type(metric)] = metric
        if log := _state_log(
            metric,
            list_items_limit=self.items_limit,
            redact_content=self.redact_content,
        ):
            ctx.log_debug(f"Recorded metric:\n⎡ {type(metric).__qualname__}:{log}\n⌊")

    async def read[Metric: State](
        self,
        scope: ScopeIdentifier,
        /,
        *,
        metric: type[Metric],
        merged: bool,
    ) -> Metric | None:
        assert self.root_scope is not None  # nosec: B101
        assert scope.scope_id in self.scopes  # nosec: B101

        if merged:
            return self.scopes[scope.scope_id].merged(metric)

        else:
            return cast(Metric | None, self.scopes[scope.scope_id].metrics.get(metric))

    def enter_scope[Metric: State](
        self,
        scope: ScopeIdentifier,
        /,
    ) -> None:
        assert scope.scope_id not in self.scopes  # nosec: B101
        scope_metrics = MetricsScopeStore(scope)
        self.scopes[scope.scope_id] = scope_metrics

        if self.root_scope is None:
            self.root_scope = scope

        else:
            for key in self.scopes.keys():
                if key == scope.parent_id:
                    self.scopes[key].nested.append(scope_metrics)
                    return

            ctx.log_debug(
                "Attempting to enter nested scope metrics without entering its parent first"
            )

    def exit_scope[Metric: State](
        self,
        scope: ScopeIdentifier,
        /,
    ) -> None:
        assert self.root_scope is not None  # nosec: B101
        assert scope.scope_id in self.scopes  # nosec: B101

        self.scopes[scope.scope_id].allow_exit = True

        if not all(nested.exited for nested in self.scopes[scope.scope_id].nested):
            return  # not completed yet

        self.scopes[scope.scope_id].exited = monotonic()

        if scope != self.root_scope and self.scopes[scope.parent_id].allow_exit:
            self.exit_scope(self.scopes[scope.parent_id].identifier)

        elif scope == self.root_scope and self.scopes[self.root_scope.scope_id].finished:
            if log := _tree_log(
                self.scopes[scope.scope_id],
                list_items_limit=self.items_limit,
                redact_content=self.redact_content,
            ):
                ctx.log_debug(f"Metrics summary:\n{log}")


def _tree_log(
    metrics: MetricsScopeStore,
    list_items_limit: int | None,
    redact_content: bool,
) -> str:
    log: str = (
        f"⎡ @{metrics.identifier.label} [{metrics.identifier.scope_id}]({metrics.time:.2f}s):"
    )

    for metric in metrics.merged():
        if type(metric) not in metrics.metrics:
            continue  # skip metrics not available in this scope

        metric_log: str = ""
        for key, value in vars(metric).items():
            if value_log := _value_log(
                value,
                list_items_limit=list_items_limit,
                redact_content=redact_content,
            ):
                metric_log += f"\n├ {key}: {value_log}"

            else:
                continue  # skip empty values

        if not metric_log:
            continue  # skip empty logs

        log += f"\n⎡ •{type(metric).__qualname__}:{metric_log.replace('\n', '\n|  ')}\n⌊"

    for nested in metrics.nested:
        nested_log: str = _tree_log(
            nested,
            list_items_limit=list_items_limit,
            redact_content=redact_content,
        )

        log += f"\n\n{nested_log}"

    return log.strip().replace("\n", "\n|  ") + "\n⌊"


def _state_log(
    value: State,
    /,
    list_items_limit: int | None,
    redact_content: bool,
) -> str | None:
    state_log: str = ""
    for key, element in vars(value).items():
        element_log: str | None = _value_log(
            element,
            list_items_limit=list_items_limit,
            redact_content=redact_content,
        )

        if element_log:
            state_log += f"\n├ {key}: {element_log}"

        else:
            continue  # skip empty logs

    if state_log:
        return state_log

    else:
        return None  # skip empty logs


def _dict_log(
    value: dict[Any, Any],
    /,
    list_items_limit: int | None,
    redact_content: bool,
) -> str | None:
    dict_log: str = ""
    for key, element in value.items():
        element_log: str | None = _value_log(
            element,
            list_items_limit=list_items_limit,
            redact_content=redact_content,
        )
        if element_log:
            dict_log += f"\n[{key}]: {element_log}"

        else:
            continue  # skip empty logs

    if dict_log:
        return dict_log.replace("\n", "\n|  ")

    else:
        return None  # skip empty logs


def _list_log(
    value: list[Any],
    /,
    list_items_limit: int | None,
    redact_content: bool,
) -> str | None:
    list_log: str = ""
    enumerated: list[tuple[int, Any]] = list(enumerate(value))
    if list_items_limit:
        if list_items_limit > 0:
            enumerated = enumerated[:list_items_limit]

        else:
            enumerated = enumerated[list_items_limit:]

    for idx, element in enumerated:
        element_log: str | None = _value_log(
            element,
            list_items_limit=list_items_limit,
            redact_content=redact_content,
        )
        if element_log:
            list_log += f"\n[{idx}] {element_log}"

        else:
            continue  # skip empty logs

    if list_log:
        return list_log.replace("\n", "\n|  ")

    else:
        return None  # skip empty logs


def _raw_value_log(
    value: Any,
    /,
    redact_content: bool,
) -> str | None:
    if value is MISSING:
        return None  # skip missing

    if redact_content:
        return "[redacted]"

    elif isinstance(value, str):
        return f'"{value}"'.replace("\n", "\n|  ")

    else:
        return str(value).strip().replace("\n", "\n|  ")


def _value_log(
    value: Any,
    /,
    list_items_limit: int | None,
    redact_content: bool,
) -> str | None:
    # try unpack dicts
    if isinstance(value, dict):
        return _dict_log(
            cast(dict[Any, Any], value),
            list_items_limit=list_items_limit,
            redact_content=redact_content,
        )

    # try unpack lists
    elif isinstance(value, list):
        return _list_log(
            cast(list[Any], value),
            list_items_limit=list_items_limit,
            redact_content=redact_content,
        )

    # try unpack state
    elif isinstance(value, State):
        return _state_log(
            value,
            list_items_limit=list_items_limit,
            redact_content=redact_content,
        )

    else:
        return _raw_value_log(
            value,
            redact_content=redact_content,
        )



from haiway.helpers.asynchrony import asynchronous, wrap_async
from haiway.helpers.caching import CacheMakeKey, CacheRead, CacheWrite, cache
from haiway.helpers.metrics import MetricsHolder, MetricsLogger
from haiway.helpers.retries import retry
from haiway.helpers.throttling import throttle
from haiway.helpers.timeouted import timeout
from haiway.helpers.tracing import ArgumentsTrace, ResultTrace, traced

__all__ = [
    "ArgumentsTrace",
    "CacheMakeKey",
    "CacheRead",
    "CacheWrite",
    "MetricsHolder",
    "MetricsLogger",
    "ResultTrace",
    "asynchronous",
    "cache",
    "retry",
    "throttle",
    "timeout",
    "traced",
    "wrap_async",
]



from asyncio import run

from haiway import ctx, setup_logging

from integrations.postgres import PostgresConnectionPool
from migrations.postgres import execute_postgres_migrations


async def migrate_databases() -> None:
    async with ctx.scope(
        "migrations",
        disposables=[
            PostgresConnectionPool(),
        ],
    ):
        ctx.log_warning("Running postgres migrations...")
        await execute_postgres_migrations()
        ctx.log_info("...postgres migrations completed")


def main() -> None:
    setup_logging("migrations")
    run(migrate_databases())


main()



from collections.abc import Callable, Sequence, Set
from copy import copy, deepcopy
from datetime import date, datetime
from enum import StrEnum
from typing import Any, Literal, NotRequired, Protocol, Required, Self, TypedDict, runtime_checkable
from uuid import UUID, uuid4

from pytest import raises

from haiway import MISSING, Default, Missing, State, frozenlist


def test_basic_initializes_with_arguments() -> None:
    class Selection(StrEnum):
        A = "A"
        B = "B"

    class Nes(State):
        val: str

    class TypedValues(TypedDict):
        val: str
        mis: int | Missing
        req: Required[Nes]
        nreq: NotRequired[bool]

    @runtime_checkable
    class Proto(Protocol):
        def __call__(self) -> None: ...

    class Basics(State):
        uuid: UUID
        date: date
        datetime: datetime
        string: str
        literal: Literal["A", "B"]
        sequence: Sequence[str]
        string_set: Set[str]
        frozen: frozenlist[int]
        integer: int
        union: str | int
        optional: str | None
        none: None
        function: Callable[[], None]
        proto: Proto
        selection: Selection
        typeddict: TypedValues

    basic = Basics(
        uuid=uuid4(),
        date=date.today(),
        datetime=datetime.now(),
        string="string",
        literal="A",
        sequence=["a", "b", "c"],
        string_set={"a", "b"},
        frozen=(1, 2, 3),
        integer=0,
        union="union",
        optional="optional",
        none=None,
        function=lambda: None,
        proto=lambda: None,
        selection=Selection.A,
        typeddict={
            "val": "ok",
            "mis": 42,
            "req": Nes(val="ok"),
        },
    )
    assert basic.string == "string"
    assert basic.literal == "A"
    assert basic.sequence == ("a", "b", "c")
    assert basic.string_set == {"a", "b"}
    assert basic.frozen == (1, 2, 3)
    assert basic.integer == 0
    assert basic.union == "union"
    assert basic.optional == "optional"
    assert basic.none is None
    assert basic.selection == Selection.A
    assert basic.typeddict == TypedValues(
        val="ok",
        mis=42,
        req=Nes(val="ok"),
    )
    assert callable(basic.function)
    assert isinstance(basic.proto, Proto)


def test_basic_initializes_with_defaults() -> None:
    class Basics(State):
        string: str = ""
        integer: int = 0
        optional: str | None = None
        unique: UUID = Default(factory=uuid4)
        same: UUID = Default(uuid4())

    basic = Basics()
    assert basic.string == ""
    assert basic.integer == 0
    assert basic.optional is None
    assert basic.unique is not Basics().unique
    assert basic.same is Basics().same


def test_basic_equals_checks_properties() -> None:
    class Basics(State):
        string: str
        integer: int

    assert Basics(string="a", integer=1) == Basics(string="a", integer=1)
    assert Basics(string="a", integer=1) != Basics(string="b", integer=1)
    assert Basics(string="a", integer=1) != Basics(string="a", integer=2)
    assert Basics(string="a", integer=1) != Basics(string="b", integer=2)


def test_basic_initializes_with_arguments_and_defaults() -> None:
    class Basics(State):
        string: str
        integer: int = 0
        optional: str | None = None

    basic = Basics(
        string="string",
        integer=42,
    )
    assert basic.string == "string"
    assert basic.integer == 42
    assert basic.optional is None


def test_parametrized_initializes_with_proper_parameters() -> None:
    class Parametrized[T](State):
        value: T

    parametrized_string = Parametrized(
        value="string",
    )
    assert parametrized_string.value == "string"

    parametrized_int = Parametrized(
        value=42,
    )
    assert parametrized_int.value == 42

    assert parametrized_string != parametrized_int


def test_nested_initializes_with_proper_arguments() -> None:
    class Nested(State):
        string: str

    class Recursive(State):
        nested: Nested
        recursion: "Recursive | None"
        self_recursion: Self | None

    recursive = Recursive(
        nested=Nested(string="one"),
        recursion=Recursive(
            nested=Nested(string="two"),
            recursion=None,
            self_recursion=None,
        ),
        self_recursion=None,
    )
    assert recursive.nested == Nested(string="one")
    assert recursive.recursion == Recursive(
        nested=Nested(string="two"),
        recursion=None,
        self_recursion=None,
    )


def test_dict_skips_missing_properties() -> None:
    class Basics(State):
        string: str
        integer: int | Missing | None

    assert Basics(string="a", integer=1).as_dict() == {"string": "a", "integer": 1}
    assert Basics(string="a", integer=MISSING).as_dict() == {"string": "a"}
    assert Basics(string="a", integer=None).as_dict() == {"string": "a", "integer": None}


def test_initialization_allows_missing_properties() -> None:
    class Basics(State):
        string: str
        integer: int | Missing | None

    assert Basics(**{"string": "a", "integer": 1}) == Basics(string="a", integer=1)
    assert Basics(**{"string": "a", "integer": None}) == Basics(string="a", integer=None)
    assert Basics(**{"string": "a"}) == Basics(string="a", integer=MISSING)


def test_generic_subtypes_validation() -> None:
    class NestedGeneric[T](State):
        value: T

    class Generic[T](State):
        nested: NestedGeneric[T]

    class Container(State):
        generic: Generic[str]

    assert isinstance(Generic[str](nested=NestedGeneric[str](value="ok")), Generic)
    assert isinstance(Generic(nested=NestedGeneric[str](value="ok")), Generic)
    assert isinstance(Generic(nested=NestedGeneric(value="ok")), Generic)

    assert isinstance(Generic[str](nested=NestedGeneric[str](value="ok")), Generic[Any])
    assert isinstance(Generic(nested=NestedGeneric[str](value="ok")), Generic[Any])
    assert isinstance(Generic[str](nested=NestedGeneric(value="ok")), Generic[Any])
    assert isinstance(Generic(nested=NestedGeneric(value="ok")), Generic[Any])

    assert isinstance(Generic[str](nested=NestedGeneric[str](value="ok")), Generic[str])
    assert isinstance(Generic(nested=NestedGeneric[str](value="ok")), Generic[str])
    assert isinstance(Generic[str](nested=NestedGeneric(value="ok")), Generic[str])
    assert isinstance(Generic(nested=NestedGeneric(value="ok")), Generic[str])

    with raises(TypeError):
        _ = Generic[int](nested=NestedGeneric[str](value="ok"))

    with raises(TypeError):
        _ = Container(generic=Generic(nested=NestedGeneric(value=42)))

    with raises(TypeError):
        _ = Container(generic=Generic[int](nested=NestedGeneric[str](value="ok")))

    # not raises
    _ = Container(generic=Generic(nested=NestedGeneric(value="ok")))


def test_copying_leaves_same_object() -> None:
    class Nested(State):
        string: str

    class Copied(State):
        string: str
        nested: Nested

    origin = Copied(string="42", nested=Nested(string="answer"))
    assert copy(origin) is origin
    assert deepcopy(origin) is origin



from asyncio import AbstractEventLoop, get_running_loop, iscoroutinefunction
from collections.abc import Callable, Coroutine
from concurrent.futures import Executor
from contextvars import Context, copy_context
from functools import partial
from typing import Any, cast, overload

from haiway.types.missing import MISSING, Missing

__all__ = [
    "asynchronous",
    "wrap_async",
]


def wrap_async[**Args, Result](
    function: Callable[Args, Coroutine[Any, Any, Result]] | Callable[Args, Result],
    /,
) -> Callable[Args, Coroutine[Any, Any, Result]]:
    if iscoroutinefunction(function):
        return function

    else:

        async def async_function(*args: Args.args, **kwargs: Args.kwargs) -> Result:
            return cast(Callable[Args, Result], function)(*args, **kwargs)

        _mimic_async(function, within=async_function)
        return async_function


@overload
def asynchronous[**Args, Result]() -> Callable[
    [Callable[Args, Result]],
    Callable[Args, Coroutine[Any, Any, Result]],
]: ...


@overload
def asynchronous[**Args, Result](
    *,
    loop: AbstractEventLoop | None = None,
    executor: Executor,
) -> Callable[
    [Callable[Args, Result]],
    Callable[Args, Coroutine[Any, Any, Result]],
]: ...


@overload
def asynchronous[**Args, Result](
    function: Callable[Args, Result],
    /,
) -> Callable[Args, Coroutine[Any, Any, Result]]: ...


def asynchronous[**Args, Result](
    function: Callable[Args, Result] | None = None,
    /,
    loop: AbstractEventLoop | None = None,
    executor: Executor | Missing = MISSING,
) -> (
    Callable[
        [Callable[Args, Result]],
        Callable[Args, Coroutine[Any, Any, Result]],
    ]
    | Callable[Args, Coroutine[Any, Any, Result]]
):
    """\
    Wrapper for a sync function to convert it to an async function. \
    When specified an executor, it can be used to wrap long running or blocking synchronous \
    operations within coroutines system.

    Parameters
    ----------
    function: Callable[Args, Result]
        function to be wrapped as running in loop executor.
    loop: AbstractEventLoop | None
        loop used to call the function. When None was provided the loop currently running while \
        executing the function will be used. Default is None.
    executor: Executor | Missing
        executor used to run the function. When not provided (Missing) default loop executor\
         will be used.

    Returns
    -------
    Callable[_Args, _Result]
        function wrapped to async using loop executor.
    """

    def wrap(
        wrapped: Callable[Args, Result],
    ) -> Callable[Args, Coroutine[Any, Any, Result]]:
        assert not iscoroutinefunction(wrapped), "Cannot wrap async function in executor"  # nosec: B101

        return _ExecutorWrapper(
            wrapped,
            loop=loop,
            executor=cast(Executor | None, None if executor is MISSING else executor),
        )

    if function := function:
        return wrap(wrapped=function)

    else:
        return wrap


class _ExecutorWrapper[**Args, Result]:
    __slots__ = (
        "__annotations__",
        "__defaults__",
        "__doc__",
        "__globals__",
        "__kwdefaults__",
        "__name__",
        "__qualname__",
        "__wrapped__",
        "_executor",
        "_function",
        "_loop",
    )

    def __init__(
        self,
        function: Callable[Args, Result],
        /,
        loop: AbstractEventLoop | None,
        executor: Executor | None,
    ) -> None:
        self._function: Callable[Args, Result] = function
        self._loop: AbstractEventLoop | None = loop
        self._executor: Executor | None = executor

        # mimic function attributes if able
        _mimic_async(function, within=self)

    async def __call__(
        self,
        *args: Args.args,
        **kwargs: Args.kwargs,
    ) -> Result:
        context: Context = copy_context()
        return await (self._loop or get_running_loop()).run_in_executor(
            self._executor,
            context.run,
            partial(self._function, *args, **kwargs),
        )

    def __get__(
        self,
        instance: object,
        owner: type | None = None,
        /,
    ) -> Callable[Args, Coroutine[Any, Any, Result]]:
        if owner is None:
            return self

        else:
            return _mimic_async(
                self._function,
                within=partial(
                    self.__method_call__,
                    instance,
                ),
            )

    async def __method_call__(
        self,
        __method_self: object,
        *args: Args.args,
        **kwargs: Args.kwargs,
    ) -> Result:
        return await (self._loop or get_running_loop()).run_in_executor(
            self._executor,
            partial(self._function, __method_self, *args, **kwargs),
        )


def _mimic_async[**Args, Result](
    function: Callable[Args, Result],
    /,
    within: Callable[..., Coroutine[Any, Any, Result]],
) -> Callable[Args, Coroutine[Any, Any, Result]]:
    try:
        annotations: Any = getattr(  # noqa: B009
            function,
            "__annotations__",
        )
        setattr(  # noqa: B010
            within,
            "__annotations__",
            {
                **annotations,
                "return": Coroutine[Any, Any, annotations.get("return", Any)],
            },
        )

    except AttributeError:
        pass

    for attribute in (
        "__module__",
        "__name__",
        "__qualname__",
        "__doc__",
        "__type_params__",
        "__defaults__",
        "__kwdefaults__",
        "__globals__",
    ):
        try:
            setattr(
                within,
                attribute,
                getattr(
                    function,
                    attribute,
                ),
            )

        except AttributeError:
            pass
    try:
        within.__dict__.update(function.__dict__)

    except AttributeError:
        pass

    setattr(  # noqa: B010 - mimic functools.wraps behavior for correct signature checks
        within,
        "__wrapped__",
        function,
    )

    return cast(
        Callable[Args, Coroutine[Any, Any, Result]],
        within,
    )



from asyncio import AbstractEventLoop, CancelledError, Future, get_running_loop
from collections import deque
from collections.abc import AsyncIterator
from typing import Any

__all__ = [
    "AsyncQueue",
]


class AsyncQueue[Element](AsyncIterator[Element]):
    """
    Asynchronous queue supporting iteration and finishing.
    Cannot be concurrently consumed by multiple readers.
    """

    __slots__ = (
        "_finish_reason",
        "_loop",
        "_queue",
        "_waiting",
    )

    def __init__(
        self,
        *elements: Element,
        loop: AbstractEventLoop | None = None,
    ) -> None:
        self._loop: AbstractEventLoop
        object.__setattr__(
            self,
            "_loop",
            loop or get_running_loop(),
        )
        self._queue: deque[Element]
        object.__setattr__(
            self,
            "_queue",
            deque(elements),
        )
        self._waiting: Future[Element] | None
        object.__setattr__(
            self,
            "_waiting",
            None,
        )
        self._finish_reason: BaseException | None
        object.__setattr__(
            self,
            "_finish_reason",
            None,
        )

    def __setattr__(
        self,
        name: str,
        value: Any,
    ) -> Any:
        raise AttributeError(
            f"Can't modify immutable {self.__class__.__qualname__},"
            f" attribute - '{name}' cannot be modified"
        )

    def __delattr__(
        self,
        name: str,
    ) -> None:
        raise AttributeError(
            f"Can't modify immutable {self.__class__.__qualname__},"
            f" attribute - '{name}' cannot be deleted"
        )

    @property
    def is_finished(self) -> bool:
        return self._finish_reason is not None

    def enqueue(
        self,
        element: Element,
        /,
    ) -> None:
        if self.is_finished:
            raise RuntimeError("AsyncQueue is already finished")

        if self._waiting is not None and not self._waiting.done():
            self._waiting.set_result(element)

        else:
            self._queue.append(element)

    def finish(
        self,
        exception: BaseException | None = None,
    ) -> None:
        if self.is_finished:
            return  # already finished, ignore

        object.__setattr__(
            self,
            "_finish_reason",
            exception or StopAsyncIteration(),
        )

        if self._waiting is not None and not self._waiting.done():
            # checking loop only on finish as the rest of operations
            # should always have a valid loop in a typical environment
            # and we are not supporting multithreading yet
            if get_running_loop() is not self._loop:
                self._loop.call_soon_threadsafe(
                    self._waiting.set_exception,
                    self._finish_reason,  # pyright: ignore[reportArgumentType]
                )

            else:
                self._waiting.set_exception(self._finish_reason)  # pyright: ignore[reportArgumentType]

    def cancel(self) -> None:
        self.finish(exception=CancelledError())

    async def __anext__(self) -> Element:
        assert self._waiting is None, "Only a single queue consumer is supported!"  # nosec: B101

        if self._queue:  # check the queue, let it finish
            return self._queue.popleft()

        if self._finish_reason is not None:  # check if is finished
            raise self._finish_reason

        try:
            # create a new future to wait for next
            object.__setattr__(
                self,
                "_waiting",
                self._loop.create_future(),
            )
            # wait for the result
            return await self._waiting  # pyright: ignore[reportGeneralTypeIssues]

        finally:
            # cleanup
            object.__setattr__(
                self,
                "_waiting",
                None,
            )

    def clear(self) -> None:
        if self._waiting is None or self._waiting.done():
            self._queue.clear()



from collections.abc import Iterable, Mapping, Set
from typing import overload

__all__ = [
    "as_dict",
    "as_list",
    "as_set",
    "as_tuple",
]


@overload
def as_list[T](
    collection: Iterable[T],
    /,
) -> list[T]: ...


@overload
def as_list[T](
    collection: Iterable[T] | None,
    /,
) -> list[T] | None: ...


def as_list[T](
    collection: Iterable[T] | None,
    /,
) -> list[T] | None:
    """
    Converts any given Iterable into a list.

    Parameters
    ----------
    collection : Iterable[T] | None
        The input to be converted.

    Returns
    -------
    list[T] | None
        A new list containing all elements of the input,\
         or the original list if it was already one.
        None if no value was provided.
    """

    if collection is None:
        return None

    if isinstance(collection, list):
        return collection

    else:
        return list(collection)


@overload
def as_tuple[T](
    collection: Iterable[T],
    /,
) -> tuple[T, ...]: ...


@overload
def as_tuple[T](
    collection: Iterable[T] | None,
    /,
) -> tuple[T, ...] | None: ...


def as_tuple[T](
    collection: Iterable[T] | None,
    /,
) -> tuple[T, ...] | None:
    """
    Converts any given Iterable into a tuple.

    Parameters
    ----------
    collection : Sequence[T] | None
        The input to be converted.

    Returns
    -------
    tuple[T] | None
        A new tuple containing all elements of the input,\
         or the original tuple if it was already one.
        None if no value was provided.
    """

    if collection is None:
        return None

    if isinstance(collection, tuple):
        return collection

    else:
        return tuple(collection)


@overload
def as_set[T](
    collection: Set[T],
    /,
) -> set[T]: ...


@overload
def as_set[T](
    collection: Set[T] | None,
    /,
) -> set[T] | None: ...


def as_set[T](
    collection: Set[T] | None,
    /,
) -> set[T] | None:
    """
    Converts any given Set into a set.

    Parameters
    ----------
    collection : Set[T]
        The input collection to be converted.

    Returns
    -------
    set[T]
        A new set containing all elements of the input collection,\
         or the original set if it was already one.
        None if no value was provided.
    """

    if collection is None:
        return None

    if isinstance(collection, set):
        return collection

    else:
        return set(collection)


@overload
def as_dict[K, V](
    collection: Mapping[K, V],
    /,
) -> dict[K, V]: ...


@overload
def as_dict[K, V](
    collection: Mapping[K, V] | None,
    /,
) -> dict[K, V] | None: ...


def as_dict[K, V](
    collection: Mapping[K, V] | None,
    /,
) -> dict[K, V] | None:
    """
    Converts any given Mapping into a dict.

    Parameters
    ----------
    collection : Mapping[K, V]
        The input collection to be converted.

    Returns
    -------
    dict[K, V]
        A new dict containing all elements of the input collection,\
         or the original dict if it was already one.
        None if no value was provided.
    """

    if collection is None:
        return None

    if isinstance(collection, dict):
        return collection

    else:
        return dict(collection)



from typing import Protocol, runtime_checkable
from uuid import UUID

__all__ = [
    "TodoCompletion",
]


@runtime_checkable
class TodoCompletion(Protocol):
    async def __call__(
        self,
        *,
        identifier: UUID,
    ) -> None: ...



from asyncio import Task, TaskGroup, get_event_loop
from collections.abc import Callable, Coroutine
from contextvars import ContextVar, Token, copy_context
from types import TracebackType
from typing import Any, final

__all__ = [
    "TaskGroupContext",
]


@final
class TaskGroupContext:
    _context = ContextVar[TaskGroup]("TaskGroupContext")

    @classmethod
    def run[Result, **Arguments](
        cls,
        function: Callable[Arguments, Coroutine[Any, Any, Result]],
        /,
        *args: Arguments.args,
        **kwargs: Arguments.kwargs,
    ) -> Task[Result]:
        try:
            return cls._context.get().create_task(
                function(*args, **kwargs),
                context=copy_context(),
            )

        except LookupError:  # spawn task out of group as a fallback
            return get_event_loop().create_task(
                function(*args, **kwargs),
                context=copy_context(),
            )

    __slots__ = (
        "_group",
        "_token",
    )

    def __init__(
        self,
        task_group: TaskGroup | None = None,
    ) -> None:
        self._group: TaskGroup
        object.__setattr__(
            self,
            "_group",
            task_group if task_group is not None else TaskGroup(),
        )
        self._token: Token[TaskGroup] | None
        object.__setattr__(
            self,
            "_token",
            None,
        )

    def __setattr__(
        self,
        name: str,
        value: Any,
    ) -> Any:
        raise AttributeError(
            f"Can't modify immutable {self.__class__.__qualname__},"
            f" attribute - '{name}' cannot be modified"
        )

    def __delattr__(
        self,
        name: str,
    ) -> None:
        raise AttributeError(
            f"Can't modify immutable {self.__class__.__qualname__},"
            f" attribute - '{name}' cannot be deleted"
        )

    async def __aenter__(self) -> None:
        assert self._token is None, "Context reentrance is not allowed"  # nosec: B101
        await self._group.__aenter__()
        object.__setattr__(
            self,
            "_token",
            TaskGroupContext._context.set(self._group),
        )

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        assert self._token is not None, "Unbalanced context enter/exit"  # nosec: B101
        TaskGroupContext._context.reset(self._token)
        object.__setattr__(
            self,
            "_token",
            None,
        )

        try:
            await self._group.__aexit__(
                et=exc_type,
                exc=exc_val,
                tb=exc_tb,
            )

        except BaseException:
            pass  # silence TaskGroup exceptions, if there was exception already we will get it



from asyncio import CancelledError, sleep
from collections.abc import AsyncGenerator

from pytest import mark, raises

from haiway import ctx
from haiway.state.structure import State


class FakeException(Exception):
    pass


@mark.asyncio
async def test_fails_when_generator_fails():
    async def generator(value: int) -> AsyncGenerator[int, None]:
        yield value
        raise FakeException()

    elements: int = 0
    with raises(FakeException):
        async for _ in ctx.stream(generator, 42):
            elements += 1

    assert elements == 1


@mark.asyncio
async def test_cancels_when_iteration_cancels():
    async def generator(value: int) -> AsyncGenerator[int, None]:
        await sleep(0)
        yield value

    elements: int = 0
    with raises(CancelledError):
        ctx.cancel()
        async for _ in ctx.stream(generator, 42):
            elements += 1

    assert elements == 0


@mark.asyncio
async def test_ends_when_generator_ends():
    async def generator(value: int) -> AsyncGenerator[int, None]:
        yield value

    elements: int = 0
    async for _ in ctx.stream(generator, 42):
        elements += 1

    assert elements == 1


@mark.asyncio
async def test_delivers_updates_when_generating():
    async def generator(value: int) -> AsyncGenerator[int, None]:
        for i in range(0, value):
            yield i

    elements: list[int] = []

    async for element in ctx.stream(generator, 10):
        elements.append(element)

    assert elements == list(range(0, 10))


@mark.asyncio
async def test_streaming_context_variables_access_is_preserved():
    class TestState(State):
        value: int = 42
        other: str = "other"

    async def generator(value: int) -> AsyncGenerator[TestState, None]:
        yield ctx.state(TestState)
        with ctx.scope("nested", ctx.state(TestState).updated(value=value)):
            yield ctx.state(TestState)

    async with ctx.scope("test", TestState(value=42)):
        elements: list[TestState] = []

        async for element in ctx.stream(generator, 10):
            elements.append(element)

    assert elements == [
        TestState(value=42),
        TestState(value=10),
    ]


@mark.asyncio
async def test_nested_streaming_streams_correctly():
    class TestState(State):
        value: int = 42
        other: str = "other"

    async def inner(value: int) -> AsyncGenerator[TestState, None]:
        yield ctx.state(TestState)
        with ctx.scope("inner", ctx.state(TestState).updated(value=value, other="inner")):
            yield ctx.state(TestState)

    async def outer(value: int) -> AsyncGenerator[TestState, None]:
        yield ctx.state(TestState)
        with ctx.scope("outer", ctx.state(TestState).updated(other="outer")):
            async for item in ctx.stream(inner, value):
                yield item

    async with ctx.scope("test", TestState(value=42)):
        elements: list[TestState] = []

        async for element in ctx.stream(outer, 10):
            elements.append(element)

    assert elements == [
        TestState(value=42),
        TestState(value=42, other="outer"),
        TestState(value=10, other="inner"),
    ]



from uuid import UUID

from fastapi import APIRouter
from haiway import ctx
from starlette.responses import Response

from features.todos import Todos
from integrations.postgres import Postgres

__all__ = [
    "router",
]

router = APIRouter()


@router.post(
    path="/todo/{identifier}/complete",
    description="Complete a TODO.",
    status_code=204,
    responses={
        204: {"description": "TODO has been completed!"},
        500: {"description": "Internal server error"},
    },
)
async def complete_todo_endpoint(identifier: UUID) -> Response:
    async with ctx.scope(
        "todo_completion",
        disposables=(Postgres.connection(),),
    ):
        await Todos.complete_todo(identifier=identifier)
        return Response(status_code=204)



from collections.abc import Sequence
from typing import overload
from uuid import UUID

from haiway import State, ctx

from solutions.user_tasks.postgres import (
    postgres_task_create,
    postgres_task_delete,
    postgres_task_update,
    postgres_tasks_fetch,
)
from solutions.user_tasks.types import (
    UserTask,
    UserTaskCreation,
    UserTaskDeletion,
    UserTaskFetching,
    UserTaskUpdating,
)

__all__ = [
    "UserTasks",
]


class UserTasks(State):
    @classmethod
    async def create_task(
        cls,
        description: str,
    ) -> UserTask:
        return await ctx.state(UserTasks).create(description=description)

    @classmethod
    async def update_task(
        cls,
        task: UserTask,
        /,
    ) -> None:
        await ctx.state(cls).update(task=task)

    @overload
    @classmethod
    async def fetch_tasks(
        cls,
        *,
        identifier: None = None,
    ) -> Sequence[UserTask]: ...

    @overload
    @classmethod
    async def fetch_tasks(
        cls,
        *,
        identifier: UUID,
    ) -> UserTask: ...

    @classmethod
    async def fetch_tasks(
        cls,
        *,
        identifier: UUID | None = None,
    ) -> Sequence[UserTask] | UserTask:
        return await ctx.state(cls).fetch(identifier=identifier)

    @classmethod
    async def delete_task(
        cls,
        *,
        identifier: UUID,
    ) -> None:
        return await ctx.state(cls).delete(identifier=identifier)

    fetch: UserTaskFetching = postgres_tasks_fetch
    create: UserTaskCreation = postgres_task_create
    update: UserTaskUpdating = postgres_task_update
    delete: UserTaskDeletion = postgres_task_delete



from collections.abc import Sequence
from datetime import datetime
from typing import overload
from uuid import UUID

from haiway import ctx

from integrations.postgres import PostgresConnection, PostgresException
from solutions.user_tasks.types import UserTask

__all__ = [
    "postgres_task_create",
    "postgres_task_delete",
    "postgres_task_update",
    "postgres_tasks_fetch",
]

CREATE_TODO_STATEMENT: str = """\
INSERT INTO
    todos (
        description
    )

VALUES
    (
        $1
    )

RETURNING
    id,
    modified,
    description,
    completed
;
"""


async def postgres_task_create(
    *,
    description: str,
) -> UserTask:
    try:  # actual SQL goes here...
        match await PostgresConnection.execute(
            CREATE_TODO_STATEMENT,
            description,
        ):
            case {
                "id": str() as identifier,
                "description": str() as description,
                "modified": datetime() as modified,
                "completed": bool() as completed,
            }:
                return UserTask(
                    identifier=identifier,
                    description=description,
                    modified=modified,
                    completed=completed,
                )

            case other:
                raise ValueError(f"Invalid database record:\n{other}")

    except PostgresException as exc:
        ctx.log_debug(
            "postgres_task_create failed",
            exception=exc,
        )
        raise exc


UPDATE_TODO_STATEMENT: str = """\
UPDATE
    todos

SET
    modified = now(),
    description = $2,
    completed =$3

WHERE
    id = $1
;
"""


async def postgres_task_update(
    *,
    task: UserTask,
) -> None:
    try:  # actual SQL goes here...
        await PostgresConnection.execute(
            UPDATE_TODO_STATEMENT,
            task.identifier,
            task.description,
            task.completed,
        )

    except PostgresException as exc:
        ctx.log_debug(
            "postgres_task_update failed",
            exception=exc,
        )
        raise exc


@overload
async def postgres_tasks_fetch(
    *,
    identifier: None = None,
) -> Sequence[UserTask]: ...


@overload
async def postgres_tasks_fetch(
    *,
    identifier: UUID,
) -> UserTask: ...


FETCH_TODOS_STATEMENT: str = """\
SELECT
    id,
    modified,
    description,
    completed

FROM
    todos
;
"""

FETCH_TODO_STATEMENT: str = """\
SELECT
    modified,
    description,
    completed

FROM
    todos

WHERE
    id = $1

LIMIT
    1
;
"""


async def postgres_tasks_fetch(
    identifier: UUID | None = None,
) -> Sequence[UserTask] | UserTask | None:
    try:  # actual SQL goes here...
        if identifier:
            match await PostgresConnection.fetch_one(FETCH_TODO_STATEMENT, identifier):
                case None:
                    return None

                case {
                    "modified": datetime() as modified,
                    "description": str() as description,
                    "completed": bool() as completed,
                }:
                    return UserTask(
                        identifier=identifier,
                        description=description,
                        modified=modified,
                        completed=completed,
                    )

                case other:
                    raise ValueError(f"Invalid database record:\n{other}")

        else:
            return [
                UserTask(
                    identifier=task["id"],
                    description=task["description"],
                    modified=task["modified"],
                    completed=task["completed"],
                )
                for task in await PostgresConnection.fetch(FETCH_TODOS_STATEMENT)
            ]

    except PostgresException as exc:
        ctx.log_debug(
            "postgres_tasks_fetch failed",
            exception=exc,
        )
        raise exc


DELETE_TODO_STATEMENT: str = """\
DELETE
    todos

WHERE
    id = $1
;
"""


async def postgres_task_delete(
    *,
    identifier: UUID,
) -> None:
    try:  # actual SQL goes here...
        await PostgresConnection.execute(
            DELETE_TODO_STATEMENT,
            identifier,
        )

    except PostgresException as exc:
        ctx.log_debug(
            "postgres_task_delete failed",
            exception=exc,
        )
        raise exc



from uuid import UUID

from solutions.user_tasks import UserTasks

__all__ = [
    "complete_todo_task",
]


async def complete_todo_task(
    *,
    identifier: UUID,
) -> None:
    match await UserTasks.fetch(identifier=identifier):
        case None:
            pass

        case task:
            await UserTasks.update(task=task.updated(completed=True))



from asyncio import iscoroutinefunction
from collections.abc import Callable, Coroutine, Iterable, MutableMapping
from contextvars import ContextVar, Token
from types import TracebackType
from typing import Any, Self, cast, final, overload

from haiway.context.types import MissingContext, MissingState
from haiway.state import State
from haiway.utils.mimic import mimic_function

__all__ = [
    "ScopeState",
    "StateContext",
]


@final
class ScopeState:
    __slots__ = ("_state",)

    def __init__(
        self,
        state: Iterable[State],
    ) -> None:
        self._state: MutableMapping[type[State], State]
        object.__setattr__(
            self,
            "_state",
            {type(element): element for element in state},
        )

    def __setattr__(
        self,
        name: str,
        value: Any,
    ) -> Any:
        raise AttributeError(
            f"Can't modify immutable {self.__class__.__qualname__},"
            f" attribute - '{name}' cannot be modified"
        )

    def __delattr__(
        self,
        name: str,
    ) -> None:
        raise AttributeError(
            f"Can't modify immutable {self.__class__.__qualname__},"
            f" attribute - '{name}' cannot be deleted"
        )

    def state[StateType: State](
        self,
        state: type[StateType],
        /,
        default: StateType | None = None,
    ) -> StateType:
        if state in self._state:
            return cast(StateType, self._state[state])

        elif default is not None:
            return default

        else:
            try:
                initialized: StateType = state()
                # we would need a locking here in multithreaded environment
                self._state[state] = initialized
                return initialized

            except Exception as exc:
                raise MissingState(
                    f"{state.__qualname__} is not defined in current scope"
                    " and failed to provide a default value"
                ) from exc

    def updated(
        self,
        state: Iterable[State],
    ) -> Self:
        if state:
            return self.__class__(
                [
                    *self._state.values(),
                    *state,
                ]
            )

        else:
            return self


@final
class StateContext:
    _context = ContextVar[ScopeState]("StateContext")

    @classmethod
    def state[StateType: State](
        cls,
        state: type[StateType],
        /,
        default: StateType | None = None,
    ) -> StateType:
        try:
            return cls._context.get().state(state, default=default)

        except LookupError as exc:
            raise MissingContext("StateContext requested but not defined!") from exc

    @classmethod
    def updated(
        cls,
        state: Iterable[State],
        /,
    ) -> Self:
        try:
            # update current scope context
            return cls(state=cls._context.get().updated(state=state))

        except LookupError:  # create root scope when missing
            return cls(state=ScopeState(state))

    __slots__ = (
        "_state",
        "_token",
    )

    def __init__(
        self,
        state: ScopeState,
    ) -> None:
        self._state: ScopeState
        object.__setattr__(
            self,
            "_state",
            state,
        )
        self._token: Token[ScopeState] | None
        object.__setattr__(
            self,
            "_token",
            None,
        )

    def __setattr__(
        self,
        name: str,
        value: Any,
    ) -> Any:
        raise AttributeError(
            f"Can't modify immutable {self.__class__.__qualname__},"
            f" attribute - '{name}' cannot be modified"
        )

    def __delattr__(
        self,
        name: str,
    ) -> None:
        raise AttributeError(
            f"Can't modify immutable {self.__class__.__qualname__},"
            f" attribute - '{name}' cannot be deleted"
        )

    def __enter__(self) -> None:
        assert self._token is None, "Context reentrance is not allowed"  # nosec: B101
        object.__setattr__(
            self,
            "_token",
            StateContext._context.set(self._state),
        )

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        assert self._token is not None, "Unbalanced context enter/exit"  # nosec: B101
        StateContext._context.reset(self._token)
        object.__setattr__(
            self,
            "_token",
            None,
        )

    @overload
    def __call__[Result, **Arguments](
        self,
        function: Callable[Arguments, Coroutine[Any, Any, Result]],
    ) -> Callable[Arguments, Coroutine[Any, Any, Result]]: ...

    @overload
    def __call__[Result, **Arguments](
        self,
        function: Callable[Arguments, Result],
    ) -> Callable[Arguments, Result]: ...

    def __call__[Result, **Arguments](
        self,
        function: Callable[Arguments, Coroutine[Any, Any, Result]] | Callable[Arguments, Result],
    ) -> Callable[Arguments, Coroutine[Any, Any, Result]] | Callable[Arguments, Result]:
        if iscoroutinefunction(function):

            async def async_context(
                *args: Arguments.args,
                **kwargs: Arguments.kwargs,
            ) -> Result:
                with self:
                    return await function(*args, **kwargs)

            return mimic_function(function, within=async_context)

        else:

            def sync_context(
                *args: Arguments.args,
                **kwargs: Arguments.kwargs,
            ) -> Result:
                with self:
                    return function(*args, **kwargs)  # pyright: ignore[reportReturnType]

            return mimic_function(function, within=sync_context)  # pyright: ignore[reportReturnType]



from logging.config import dictConfig

from haiway.utils.env import getenv_bool

__all__ = [
    "setup_logging",
]


def setup_logging(
    *loggers: str,
    debug: bool = getenv_bool("DEBUG_LOGGING", __debug__),
) -> None:
    """\
    Setup logging configuration and prepare specified loggers.

    Parameters
    ----------
    *loggers: str
        names of additional loggers to configure.

    NOTE: this function should be run only once on application start
    """

    dictConfig(
        config={
            "version": 1,
            "disable_existing_loggers": True,
            "formatters": {
                "standard": {
                    "format": "%(asctime)s [%(levelname)-4s] [%(name)s] %(message)s",
                    "datefmt": "%d/%b/%Y:%H:%M:%S +0000",
                },
            },
            "handlers": {
                "console": {
                    "level": "DEBUG" if debug else "INFO",
                    "formatter": "standard",
                    "class": "logging.StreamHandler",
                    "stream": "ext://sys.stdout",
                },
            },
            "loggers": {
                name: {
                    "handlers": ["console"],
                    "level": "DEBUG" if debug else "INFO",
                    "propagate": False,
                }
                for name in loggers
            },
            "root": {  # root logger
                "handlers": ["console"],
                "level": "DEBUG" if debug else "INFO",
                "propagate": False,
            },
        },
    )



from typing import Final

from haiway import getenv_int, getenv_str

__all__ = [
    "SERVER_HOST",
    "SERVER_PORT",
]

SERVER_HOST: Final[str] = getenv_str("SERVER_HOST", default="localhost")
SERVER_PORT: Final[int] = getenv_int("SERVER_PORT", default=8080)



from asyncio import (
    AbstractEventLoop,
    CancelledError,
    Future,
    get_running_loop,
)
from collections.abc import AsyncIterator

__all__ = [
    "AsyncStream",
]


class AsyncStream[Element](AsyncIterator[Element]):
    def __init__(
        self,
        loop: AbstractEventLoop | None = None,
    ) -> None:
        self._loop: AbstractEventLoop = loop or get_running_loop()
        self._ready: Future[None] = self._loop.create_future()
        self._waiting: Future[Element] | None = None
        self._finish_reason: BaseException | None = None

    @property
    def finished(self) -> bool:
        return self._finish_reason is not None

    async def send(
        self,
        element: Element,
        /,
    ) -> None:
        if self._finish_reason is not None:
            return  # already finished

        # wait for readiness
        await self._ready
        # we could finish while waiting
        if self._finish_reason is not None:
            return  # already finished

        assert self._waiting is not None and not self._waiting.done()  # nosec: B101
        # send the element
        self._waiting.set_result(element)
        # and create new readiness future afterwards
        self._ready = self._loop.create_future()

    def finish(
        self,
        exception: BaseException | None = None,
    ) -> None:
        if self.finished:
            return  # already finished, ignore

        self._finish_reason = exception or StopAsyncIteration()

        if not self._ready.done():
            if get_running_loop() is not self._loop:
                self._loop.call_soon_threadsafe(
                    self._ready.set_result,
                    None,
                )

            else:
                self._ready.set_result(None)

        if self._waiting is not None and not self._waiting.done():
            if get_running_loop() is not self._loop:
                self._loop.call_soon_threadsafe(
                    self._waiting.set_exception,
                    self._finish_reason,
                )

            else:
                self._waiting.set_exception(self._finish_reason)

    def cancel(self) -> None:
        self.finish(exception=CancelledError())

    async def __anext__(self) -> Element:
        assert self._waiting is None, "AsyncStream can't be reused"  # nosec: B101

        if self._finish_reason:
            raise self._finish_reason

        try:
            assert not self._ready.done()  # nosec: B101
            # create new waiting future
            self._waiting = self._loop.create_future()
            # and notify readiness
            self._ready.set_result(None)
            # and wait for the result
            return await self._waiting

        finally:
            # cleanup waiting future
            self._waiting = None



import builtins
import types
import typing
from abc import ABC, abstractmethod
from collections import abc as collections_abc
from collections import deque
from collections.abc import Callable, Mapping, Sequence
from copy import copy
from typing import Any, TypeAliasType, final, get_args, get_origin, overload

from haiway.types import MISSING, Missing, not_missing

__all__ = [
    "AttributePath",
]


class AttributePathComponent(ABC):
    @abstractmethod
    def path_str(
        self,
        current: str | None = None,
        /,
    ) -> str: ...

    @abstractmethod
    def access(
        self,
        subject: Any,
        /,
    ) -> Any: ...

    @abstractmethod
    def assigning(
        self,
        subject: Any,
        /,
        value: Any,
    ) -> Any: ...


@final
class PropertyAttributePathComponent(AttributePathComponent):
    __slots__ = (
        "_access",
        "_assigning",
        "_name",
    )

    def __init__[Root, Attribute](
        self,
        root: type[Root],
        *,
        attribute: type[Attribute],
        name: str,
    ) -> None:
        root_origin: Any = _unaliased_origin(root)
        attribute_origin: Any = _unaliased_origin(attribute)

        def access(
            subject: Root,
            /,
        ) -> Attribute:
            assert isinstance(subject, root_origin), (  # nosec: B101
                f"AttributePath used on unexpected subject of"
                f" '{type(subject).__name__}' instead of '{root.__name__}' for '{name}'"
            )

            assert hasattr(subject, name), (  # nosec: B101
                f"AttributePath pointing to attribute '{name}'"
                f" which is not available in subject '{type(subject).__name__}'"
            )

            resolved: Any = getattr(subject, name)

            assert isinstance(resolved, attribute_origin), (  # nosec: B101
                f"AttributePath pointing to unexpected value of"
                f" '{type(resolved).__name__}' instead of '{attribute.__name__}' for '{name}'"
            )
            return resolved

        def assigning(
            subject: Root,
            /,
            value: Attribute,
        ) -> Root:
            assert isinstance(subject, root_origin), (  # nosec: B101
                f"AttributePath used on unexpected subject of"
                f" '{type(subject).__name__}' instead of '{root.__name__}' for '{name}'"
            )

            assert hasattr(subject, name), (  # nosec: B101
                f"AttributePath pointing to attribute '{name}'"
                f" which is not available in subject '{type(subject).__name__}'"
            )

            assert isinstance(value, attribute_origin), (  # nosec: B101
                f"AttributePath assigning unexpected value of "
                f"'{type(value).__name__}' instead of '{attribute.__name__}' for '{name}'"
            )

            updated: Root
            # python 3.13 introduces __replace__, we are already implementing it for our types
            if hasattr(subject, "__replace__"):  # can't check full type here
                updated = subject.__replace__(**{name: value})  # pyright: ignore[reportUnknownMemberType, reportUnknownVariableType, reportAttributeAccessIssue]

            else:
                updated = copy(subject)
                setattr(updated, name, value)

            return updated  # pyright: ignore[reportUnknownVariableType]

        self._access: Callable[[Any], Any]
        object.__setattr__(
            self,
            "_access",
            access,
        )
        self._assigning: Callable[[Any, Any], Any]
        object.__setattr__(
            self,
            "_assigning",
            assigning,
        )
        self._name: str
        object.__setattr__(
            self,
            "_name",
            name,
        )

    def __setattr__(
        self,
        name: str,
        value: Any,
    ) -> Any:
        raise AttributeError(
            f"Can't modify immutable {self.__class__.__qualname__},"
            f" attribute - '{name}' cannot be modified"
        )

    def __delattr__(
        self,
        name: str,
    ) -> None:
        raise AttributeError(
            f"Can't modify immutable {self.__class__.__qualname__},"
            f" attribute - '{name}' cannot be deleted"
        )

    def path_str(
        self,
        current: str | None = None,
        /,
    ) -> str:
        if current:
            return f"{current}.{self._name}"

        else:
            return self._name

    def access(
        self,
        subject: Any,
        /,
    ) -> Any:
        return self._access(subject)

    def assigning(
        self,
        subject: Any,
        /,
        value: Any,
    ) -> Any:
        return self._assigning(subject, value)


@final
class SequenceItemAttributePathComponent(AttributePathComponent):
    __slots__ = (
        "_access",
        "_assigning",
        "_index",
    )

    def __init__[Root: Sequence[Any], Attribute](
        self,
        root: type[Root],
        *,
        attribute: type[Attribute],
        index: int,
    ) -> None:
        root_origin: Any = _unaliased_origin(root)
        attribute_origin: Any = _unaliased_origin(attribute)

        def access(
            subject: Root,
            /,
        ) -> Attribute:
            assert isinstance(subject, root_origin), (  # nosec: B101
                f"AttributePathComponent used on unexpected root of "
                f"'{type(root).__name__}' instead of '{root.__name__}' for '{index}'"
            )

            resolved: Any = subject.__getitem__(index)  # pyright: ignore[reportUnknownMemberType, reportAttributeAccessIssue]

            assert isinstance(resolved, attribute_origin), (  # nosec: B101
                f"AttributePath pointing to unexpected value of "
                f"'{type(resolved).__name__}' instead of '{attribute.__name__}' for '{index}'"
            )
            return resolved

        def assigning(
            subject: Root,
            /,
            value: Attribute,
        ) -> Root:
            assert isinstance(subject, root_origin), (  # nosec: B101
                f"AttributePath used on unexpected root of "
                f"'{type(subject).__name__}' instead of '{root.__name__}' for '{index}'"
            )
            assert isinstance(value, attribute_origin), (  # nosec: B101
                f"AttributePath assigning to unexpected value of "
                f"'{type(value).__name__}' instead of '{attribute.__name__}' for '{index}'"
            )

            temp_list: list[Any] = list(subject)  # pyright: ignore[reportUnknownArgumentType]
            temp_list[index] = value
            return subject.__class__(temp_list)  # pyright: ignore[reportCallIssue, reportUnknownVariableType, reportUnknownMemberType]

        self._access: Callable[[Any], Any]
        object.__setattr__(
            self,
            "_access",
            access,
        )
        self._assigning: Callable[[Any, Any], Any]
        object.__setattr__(
            self,
            "_assigning",
            assigning,
        )
        self._index: Any
        object.__setattr__(
            self,
            "_index",
            index,
        )

    def __setattr__(
        self,
        name: str,
        value: Any,
    ) -> Any:
        raise AttributeError(
            f"Can't modify immutable {self.__class__.__qualname__},"
            f" attribute - '{name}' cannot be modified"
        )

    def __delattr__(
        self,
        name: str,
    ) -> None:
        raise AttributeError(
            f"Can't modify immutable {self.__class__.__qualname__},"
            f" attribute - '{name}' cannot be deleted"
        )

    def path_str(
        self,
        current: str | None = None,
        /,
    ) -> str:
        return f"{current or ''}[{self._index}]"

    def access(
        self,
        subject: Any,
        /,
    ) -> Any:
        return self._access(subject)

    def assigning(
        self,
        subject: Any,
        /,
        value: Any,
    ) -> Any:
        return self._assigning(subject, value)


@final
class MappingItemAttributePathComponent(AttributePathComponent):
    __slots__ = (
        "_access",
        "_assigning",
        "_key",
    )

    def __init__[Root: Mapping[Any, Any], Attribute](
        self,
        root: type[Root],
        *,
        attribute: type[Attribute],
        key: str | int,
    ) -> None:
        root_origin: Any = _unaliased_origin(root)
        attribute_origin: Any = _unaliased_origin(attribute)

        def access(
            subject: Root,
            /,
        ) -> Attribute:
            assert isinstance(subject, root_origin), (  # nosec: B101
                f"AttributePathComponent used on unexpected root of "
                f"'{type(root).__name__}' instead of '{root.__name__}' for '{key}'"
            )

            resolved: Any = subject.__getitem__(key)  # pyright: ignore[reportUnknownMemberType, reportAttributeAccessIssue]

            assert isinstance(resolved, attribute_origin), (  # nosec: B101
                f"AttributePath pointing to unexpected value of "
                f"'{type(resolved).__name__}' instead of '{attribute.__name__}' for '{key}'"
            )
            return resolved

        def assigning(
            subject: Root,
            /,
            value: Attribute,
        ) -> Root:
            assert isinstance(subject, root_origin), (  # nosec: B101
                f"AttributePath used on unexpected root of "
                f"'{type(subject).__name__}' instead of '{root.__name__}' for '{key}'"
            )
            assert isinstance(value, attribute_origin), (  # nosec: B101
                f"AttributePath assigning to unexpected value of "
                f"'{type(value).__name__}' instead of '{attribute.__name__}' for '{key}'"
            )

            temp_dict: dict[Any, Any] = dict(subject)  # pyright: ignore[reportUnknownArgumentType]
            temp_dict[key] = value
            return subject.__class__(temp_dict)  # pyright: ignore[reportCallIssue, reportUnknownVariableType, reportUnknownMemberType]

        self._access: Callable[[Any], Any]
        object.__setattr__(
            self,
            "_access",
            access,
        )
        self._assigning: Callable[[Any, Any], Any]
        object.__setattr__(
            self,
            "_assigning",
            assigning,
        )
        self._key: Any
        object.__setattr__(
            self,
            "_key",
            key,
        )

    def __setattr__(
        self,
        name: str,
        value: Any,
    ) -> Any:
        raise AttributeError(
            f"Can't modify immutable {self.__class__.__qualname__},"
            f" attribute - '{name}' cannot be modified"
        )

    def __delattr__(
        self,
        name: str,
    ) -> None:
        raise AttributeError(
            f"Can't modify immutable {self.__class__.__qualname__},"
            f" attribute - '{name}' cannot be deleted"
        )

    def path_str(
        self,
        current: str | None = None,
        /,
    ) -> str:
        return f"{current or ''}[{self._key}]"

    def access(
        self,
        subject: Any,
        /,
    ) -> Any:
        return self._access(subject)

    def assigning(
        self,
        subject: Any,
        /,
        value: Any,
    ) -> Any:
        return self._assigning(subject, value)


@final
class AttributePath[Root, Attribute]:
    __slots__ = (
        "__attribute__",
        "__components__",
        "__root__",
    )

    @overload
    def __init__(
        self,
        root: type[Root],
        /,
        *,
        attribute: type[Root],
    ) -> None: ...

    @overload
    def __init__(
        self,
        root: type[Root],
        /,
        *components: AttributePathComponent,
        attribute: type[Attribute],
    ) -> None: ...

    def __init__(
        self,
        root: type[Root],
        /,
        *components: AttributePathComponent,
        attribute: type[Attribute],
    ) -> None:
        assert components or root == attribute  # nosec: B101
        self.__root__: type[Root]
        object.__setattr__(
            self,
            "__root__",
            root,
        )
        self.__attribute__: type[Attribute]
        object.__setattr__(
            self,
            "__attribute__",
            attribute,
        )
        self.__components__: tuple[AttributePathComponent, ...]
        object.__setattr__(
            self,
            "__components__",
            components,
        )

    def __setattr__(
        self,
        name: str,
        value: Any,
    ) -> Any:
        raise AttributeError(
            f"Can't modify immutable {self.__class__.__qualname__},"
            f" attribute - '{name}' cannot be modified"
        )

    def __delattr__(
        self,
        name: str,
    ) -> None:
        raise AttributeError(
            f"Can't modify immutable {self.__class__.__qualname__},"
            f" attribute - '{name}' cannot be deleted"
        )

    @property
    def components(self) -> Sequence[str]:
        return tuple(component.path_str() for component in self.__components__)

    def __str__(self) -> str:
        path: str = ""
        for component in self.__components__:
            path = component.path_str(path)

        return path

    def __repr__(self) -> str:
        path: str = self.__root__.__name__
        for component in self.__components__:
            path = component.path_str(path)

        return path

    def __getattr__(
        self,
        name: str,
    ) -> Any:
        try:
            return object.__getattribute__(self, name)

        except (AttributeError, KeyError):
            pass  # continue

        assert not name.startswith(  # nosec: B101
            "_"
        ), f"Accessing private/special attribute paths ({name}) is forbidden"

        try:
            annotation: Any = self.__attribute__.__annotations__[name]

        except Exception as exc:
            raise AttributeError(
                f"Failed to prepare AttributePath caused by inaccessible"
                f" type annotation for '{name}' within '{self.__attribute__.__name__}'"
            ) from exc

        return AttributePath[Root, Any](
            self.__root__,
            *(
                *self.__components__,
                PropertyAttributePathComponent(
                    root=self.__attribute__,
                    attribute=annotation,
                    name=name,
                ),
            ),
            attribute=annotation,
        )

    def __getitem__(
        self,
        key: str | int,
    ) -> Any:
        match _unaliased_origin(self.__attribute__):
            case collections_abc.Mapping | typing.Mapping | builtins.dict:
                match get_args(_unaliased(self.__attribute__)):
                    case (builtins.str | builtins.int, element_annotation):
                        return AttributePath[Root, Any](
                            self.__root__,
                            *(
                                *self.__components__,
                                MappingItemAttributePathComponent(
                                    root=self.__attribute__,  # pyright: ignore[reportArgumentType]
                                    attribute=element_annotation,
                                    key=key,
                                ),
                            ),
                            attribute=element_annotation,
                        )

                    case other:
                        raise TypeError(
                            "Unsupported Mapping type annotation",
                            self.__attribute__.__name__,
                        )

            case builtins.tuple:
                if not isinstance(key, int):
                    raise TypeError(
                        "Unsupported tuple type annotation",
                        self.__attribute__.__name__,
                    )

                match get_args(_unaliased(self.__attribute__)):
                    case (element_annotation, builtins.Ellipsis | types.EllipsisType):
                        return AttributePath[Root, Any](
                            self.__root__,
                            *(
                                *self.__components__,
                                SequenceItemAttributePathComponent(
                                    root=self.__attribute__,  # pyright: ignore[reportArgumentType]
                                    attribute=element_annotation,
                                    index=key,
                                ),
                            ),
                            attribute=element_annotation,
                        )

                    case other:
                        return AttributePath[Root, Any](
                            self.__root__,
                            *(
                                *self.__components__,
                                SequenceItemAttributePathComponent(
                                    root=self.__attribute__,  # pyright: ignore[reportArgumentType]
                                    attribute=other[key],
                                    index=key,
                                ),
                            ),
                            attribute=other[key],
                        )

            case collections_abc.Sequence | typing.Sequence | builtins.list:
                if not isinstance(key, int):
                    raise TypeError(
                        "Unsupported Sequence type annotation",
                        self.__attribute__.__name__,
                    )

                match get_args(_unaliased(self.__attribute__)):
                    case (element_annotation,):
                        return AttributePath[Root, Any](
                            self.__root__,
                            *(
                                *self.__components__,
                                SequenceItemAttributePathComponent(
                                    root=self.__attribute__,  # pyright: ignore[reportArgumentType]
                                    attribute=element_annotation,
                                    index=key,
                                ),
                            ),
                            attribute=element_annotation,
                        )

                    case other:
                        raise TypeError(
                            "Unsupported Seqence type annotation",
                            self.__attribute__.__name__,
                        )

            case other:
                raise TypeError("Unsupported type annotation", other)

    @overload
    def __call__(
        self,
        root: Root,
        /,
    ) -> Attribute: ...

    @overload
    def __call__(
        self,
        root: Root,
        /,
        updated: Attribute,
    ) -> Root: ...

    def __call__(
        self,
        root: Root,
        /,
        updated: Attribute | Missing = MISSING,
    ) -> Root | Attribute:
        assert isinstance(root, _unaliased_origin(self.__root__)), (  # nosec: B101
            f"AttributePath '{self.__repr__()}' used on unexpected root of "
            f"'{type(root).__name__}' instead of '{self.__root__.__name__}'"
        )

        if not_missing(updated):
            assert isinstance(updated, _unaliased_origin(self.__attribute__)), (  # nosec: B101
                f"AttributePath '{self.__repr__()}' assigning to unexpected value of "
                f"'{type(updated).__name__}' instead of '{self.__attribute__.__name__}'"
            )

            resolved: Any = root
            updates_stack: deque[tuple[Any, AttributePathComponent]] = deque()
            for component in self.__components__:
                updates_stack.append((resolved, component))
                resolved = component.access(resolved)

            updated_value: Any = updated
            while updates_stack:
                subject, component = updates_stack.pop()
                updated_value = component.assigning(
                    subject,
                    value=updated_value,
                )

            return updated_value

        else:
            resolved: Any = root
            for component in self.__components__:
                resolved = component.access(resolved)

            assert isinstance(resolved, _unaliased_origin(self.__attribute__)), (  # nosec: B101
                f"AttributePath '{self.__repr__()}' pointing to unexpected value of "
                f"'{type(resolved).__name__}' instead of '{self.__attribute__.__name__}'"
            )

            return resolved


def _unaliased_origin(base: type[Any]) -> type[Any]:
    match base:
        case TypeAliasType() as aliased:
            return get_origin(aliased.__value__) or aliased.__value__

        case concrete:
            return get_origin(concrete) or concrete


def _unaliased(base: type[Any]) -> type[Any]:
    match base:
        case TypeAliasType() as aliased:
            return aliased.__value__

        case concrete:
            return concrete



from asyncio import iscoroutinefunction
from collections import OrderedDict
from collections.abc import Callable, Coroutine, Hashable
from functools import _make_key  # pyright: ignore[reportPrivateUsage]
from time import monotonic
from typing import Any, NamedTuple, Protocol, cast, overload

from haiway.context.access import ctx
from haiway.utils.mimic import mimic_function

__all__ = [
    "CacheMakeKey",
    "CacheRead",
    "CacheWrite",
    "cache",
]


class CacheMakeKey[**Args, Key](Protocol):
    def __call__(
        self,
        *args: Args.args,
        **kwargs: Args.kwargs,
    ) -> Key: ...


class CacheRead[Key, Value](Protocol):
    async def __call__(
        self,
        key: Key,
    ) -> Value | None: ...


class CacheWrite[Key, Value](Protocol):
    async def __call__(
        self,
        key: Key,
        value: Value,
    ) -> None: ...


@overload
def cache[**Args, Result](
    function: Callable[Args, Result],
    /,
) -> Callable[Args, Result]: ...


@overload
def cache[**Args, Result, Key: Hashable](
    *,
    limit: int | None = None,
    expiration: float | None = None,
    make_key: CacheMakeKey[Args, Key] | None = None,
) -> Callable[[Callable[Args, Result]], Callable[Args, Result]]: ...


@overload
def cache[**Args, Result, Key](
    *,
    make_key: CacheMakeKey[Args, Key],
    read: CacheRead[Key, Result],
    write: CacheWrite[Key, Result],
) -> Callable[
    [Callable[Args, Coroutine[Any, Any, Result]]], Callable[Args, Coroutine[Any, Any, Result]]
]: ...


def cache[**Args, Result, Key](
    function: Callable[Args, Result] | None = None,
    *,
    limit: int | None = None,
    expiration: float | None = None,
    make_key: CacheMakeKey[Args, Key] | None = None,
    read: CacheRead[Key, Result] | None = None,
    write: CacheWrite[Key, Result] | None = None,
) -> (
    Callable[
        [Callable[Args, Coroutine[Any, Any, Result]]],
        Callable[Args, Coroutine[Any, Any, Result]],
    ]
    | Callable[[Callable[Args, Result]], Callable[Args, Result]]
    | Callable[Args, Result]
):
    """
    Memoize the result of a function using a configurable cache.

    Parameters
    ----------
    function : Callable[Args, Result] | None
        The function to be memoized.
        When used as a simple decorator (i.e., `@cache`), this is the decorated function.
        Should be omitted when cache is called with configuration arguments.
    limit : int | None
        The maximum number of entries to keep in the cache.
        Defaults to 1 if not specified.
        Ignored when using custom cache implementations (read/write).
    expiration : float | None
        Time in seconds after which a cache entry expires and will be recomputed.
        Defaults to None, meaning entries don't expire based on time.
        Ignored when using custom cache implementations (read/write).
    make_key : CacheMakeKey[Args, Key] | None
        Function to generate a cache key from function arguments.
        If None, uses a default implementation that handles most cases.
        Required when using custom cache implementations (read/write).
    read : CacheRead[Key, Result] | None
        Custom asynchronous function to read values from cache.
        Must be provided together with `write` and `make_key`.
        Only available for async functions.
    write : CacheWrite[Key, Result] | None
        Custom asynchronous function to write values to cache.
        Must be provided together with `read` and `make_key`.
        Only available for async functions.

    Returns
    -------
    Callable
        If `function` is provided as a positional argument, returns the memoized function.
        Otherwise returns a decorator that can be applied to a function to memoize it
        with the given configuration.

    Notes
    -----
    This decorator supports both synchronous and asynchronous functions.
    The default implementation uses a simple in-memory LRU cache.
    For asynchronous functions, you can provide custom cache implementations
    via the `read` and `write` parameters.

    The default cache is not thread-safe and should not be used in multi-threaded
    applications without external synchronization.

    Examples
    --------
    Simple usage as a decorator:

    >>> @cache
    ... def my_function(x: int) -> int:
    ...     print("Function called")
    ...     return x * 2
    >>> my_function(5)
    Function called
    10
    >>> my_function(5)  # Cache hit, function body not executed
    10

    With configuration parameters:

    >>> @cache(limit=10, expiration=60.0)
    ... def my_function(x: int) -> int:
    ...     return x * 2

    With custom cache for async functions:

    >>> @cache(make_key=custom_key_maker, read=redis_read, write=redis_write)
    ... async def fetch_data(user_id: str) -> dict:
    ...     return await api_call(user_id)
    """

    def _wrap(function: Callable[Args, Result]) -> Callable[Args, Result]:
        if iscoroutinefunction(function):
            if read is not None and write is not None and make_key is not None:
                assert limit is None and expiration is None  # nosec: B101
                return cast(
                    Callable[Args, Result],
                    _CustomCache(
                        function,
                        make_key=make_key,
                        read=read,
                        write=write,
                    ),
                )

            else:
                assert read is None and write is None  # nosec: B101
                return cast(
                    Callable[Args, Result],
                    _AsyncCache(
                        function,
                        limit=limit if limit is not None else 1,
                        expiration=expiration,
                        make_key=cast(
                            CacheMakeKey[Args, Hashable],
                            make_key if make_key is not None else _default_make_key,
                        ),
                    ),
                )

        else:
            assert read is None and write is None, "Custom sync cache is not supported"  # nosec: B101
            return cast(
                Callable[Args, Result],
                _SyncCache(
                    function,
                    limit=limit if limit is not None else 1,
                    expiration=expiration,
                    make_key=cast(
                        CacheMakeKey[Args, Hashable],
                        make_key if make_key is not None else _default_make_key,
                    ),
                ),
            )

    if function := function:
        return _wrap(function)

    else:
        return _wrap


class _CacheEntry[Entry](NamedTuple):
    value: Entry
    expire: float | None


class _SyncCache[**Args, Result]:
    __slots__ = (
        "__annotations__",
        "__defaults__",
        "__doc__",
        "__globals__",
        "__kwdefaults__",
        "__name__",
        "__qualname__",
        "__wrapped__",
        "_cached",
        "_function",
        "_limit",
        "_make_key",
        "_next_expire_time",
    )

    def __init__(
        self,
        function: Callable[Args, Result],
        /,
        limit: int,
        expiration: float | None,
        make_key: CacheMakeKey[Args, Hashable],
    ) -> None:
        self._function: Callable[Args, Result] = function
        self._cached: OrderedDict[Hashable, _CacheEntry[Result]] = OrderedDict()
        self._limit: int = limit
        self._make_key: CacheMakeKey[Args, Hashable] = make_key

        if expiration := expiration:

            def next_expire_time() -> float | None:
                return monotonic() + expiration

        else:

            def next_expire_time() -> float | None:
                return None

        self._next_expire_time: Callable[[], float | None] = next_expire_time

        # mimic function attributes if able
        mimic_function(function, within=self)

    def __get__(
        self,
        instance: object | None,
        owner: type | None = None,
        /,
    ) -> Callable[Args, Result]:
        assert owner is None and instance is None, "cache does not work for classes"  # nosec: B101
        return self

    def __call__(
        self,
        *args: Args.args,
        **kwargs: Args.kwargs,
    ) -> Result:
        key: Hashable = self._make_key(
            *args,
            **kwargs,
        )

        match self._cached.get(key):
            case None:
                pass

            case entry:
                if (expire := entry[1]) and expire < monotonic():
                    del self._cached[key]  # continue the same way as if empty

                else:
                    self._cached.move_to_end(key)
                    return entry[0]

        result: Result = self._function(*args, **kwargs)
        self._cached[key] = _CacheEntry(
            value=result,
            expire=self._next_expire_time(),
        )

        if len(self._cached) > self._limit:
            # keep the size limit
            self._cached.popitem(last=False)

        return result


class _AsyncCache[**Args, Result]:
    __slots__ = (
        "__annotations__",
        "__defaults__",
        "__doc__",
        "__globals__",
        "__kwdefaults__",
        "__name__",
        "__qualname__",
        "__wrapped__",
        "_cached",
        "_function",
        "_limit",
        "_make_key",
        "_next_expire_time",
    )

    def __init__(
        self,
        function: Callable[Args, Coroutine[Any, Any, Result]],
        /,
        limit: int,
        expiration: float | None,
        make_key: CacheMakeKey[Args, Hashable],
    ) -> None:
        self._function: Callable[Args, Coroutine[Any, Any, Result]] = function
        self._cached: OrderedDict[Hashable, _CacheEntry[Result]] = OrderedDict()
        self._limit: int = limit
        self._make_key: CacheMakeKey[Args, Hashable] = make_key

        if expiration := expiration:

            def next_expire_time() -> float | None:
                return monotonic() + expiration

        else:

            def next_expire_time() -> float | None:
                return None

        self._next_expire_time: Callable[[], float | None] = next_expire_time

        # mimic function attributes if able
        mimic_function(function, within=self)

    def __get__(
        self,
        instance: object | None,
        owner: type | None = None,
        /,
    ) -> Callable[Args, Coroutine[Any, Any, Result]]:
        assert owner is None and instance is None, "cache does not work for classes"  # nosec: B101
        return self

    async def __call__(
        self,
        *args: Args.args,
        **kwargs: Args.kwargs,
    ) -> Result:
        key: Hashable = self._make_key(
            *args,
            **kwargs,
        )

        match self._cached.get(key):
            case None:
                pass

            case entry:
                if (expire := entry[1]) and expire < monotonic():
                    del self._cached[key]  # continue the same way as if empty

                else:
                    self._cached.move_to_end(key)
                    return entry[0]

        result: Result = await self._function(*args, **kwargs)
        self._cached[key] = _CacheEntry(
            value=result,
            expire=self._next_expire_time(),
        )
        if len(self._cached) > self._limit:
            # keep the size limit
            self._cached.popitem(last=False)

        return result


class _CustomCache[**Args, Result, Key]:
    __slots__ = (
        "__annotations__",
        "__defaults__",
        "__doc__",
        "__globals__",
        "__kwdefaults__",
        "__name__",
        "__qualname__",
        "__wrapped__",
        "_expiration",
        "_function",
        "_make_key",
        "_read",
        "_write",
    )

    def __init__(
        self,
        function: Callable[Args, Coroutine[Any, Any, Result]],
        /,
        make_key: CacheMakeKey[Args, Key],
        read: CacheRead[Key, Result],
        write: CacheWrite[Key, Result],
    ) -> None:
        self._function: Callable[Args, Coroutine[Any, Any, Result]] = function
        self._make_key: CacheMakeKey[Args, Key] = make_key
        self._read: CacheRead[Key, Result] = read
        self._write: CacheWrite[Key, Result] = write

        # mimic function attributes if able
        mimic_function(function, within=self)

    async def __call__(
        self,
        *args: Args.args,
        **kwargs: Args.kwargs,
    ) -> Result:
        key: Key = self._make_key(
            *args,
            **kwargs,
        )

        match await self._read(key):
            case None:
                result: Result = await self._function(*args, **kwargs)
                ctx.spawn(  # write the value asnychronously
                    self._write,
                    key=key,
                    value=result,
                )

                return result

            case entry:
                return entry


def _default_make_key[**Args](
    *args: Args.args,
    **kwargs: Args.kwargs,
) -> Hashable:
    return _make_key(
        args=args,
        kwds=kwargs,
        typed=True,
    )



from asyncio import (
    Lock,
    iscoroutinefunction,
    sleep,
)
from collections import deque
from collections.abc import Callable, Coroutine
from datetime import timedelta
from time import monotonic
from typing import Any, cast, overload

from haiway.utils.mimic import mimic_function

__all__ = [
    "throttle",
]


@overload
def throttle[**Args, Result](
    function: Callable[Args, Coroutine[Any, Any, Result]],
    /,
) -> Callable[Args, Coroutine[Any, Any, Result]]: ...


@overload
def throttle[**Args, Result](
    *,
    limit: int = 1,
    period: timedelta | float = 1,
) -> Callable[
    [Callable[Args, Coroutine[Any, Any, Result]]], Callable[Args, Coroutine[Any, Any, Result]]
]: ...


def throttle[**Args, Result](
    function: Callable[Args, Coroutine[Any, Any, Result]] | None = None,
    *,
    limit: int = 1,
    period: timedelta | float = 1,
) -> (
    Callable[
        [Callable[Args, Coroutine[Any, Any, Result]]],
        Callable[Args, Coroutine[Any, Any, Result]],
    ]
    | Callable[Args, Coroutine[Any, Any, Result]]
):
    """\
    Throttle for function calls with custom limit and period time. \
    Works only for async functions by waiting desired time before execution. \
    It is not allowed to be used on class or instance methods. \
    This wrapper is not thread safe.

    Parameters
    ----------
    function: Callable[Args, Coroutine[Any, Any, Result]]
        function to wrap in throttle
    limit: int
        limit of executions in given period, if no period was specified
        it is number of concurrent executions instead, default is 1
    period: timedelta | float | None
        period time (in seconds by default) during which the limit resets, default is 1 second

    Returns
    -------
    Callable[[Callable[Args, Coroutine[Any, Any, Result]]], Callable[Args, Coroutine[Any, Any, Result]]] \
    | Callable[Args, Coroutine[Any, Any, Result]]
        provided function wrapped in throttle
    """  # noqa: E501

    def _wrap(
        function: Callable[Args, Coroutine[Any, Any, Result]],
    ) -> Callable[Args, Coroutine[Any, Any, Result]]:
        assert iscoroutinefunction(function)  # nosec: B101
        return cast(
            Callable[Args, Coroutine[Any, Any, Result]],
            _AsyncThrottle(
                function,
                limit=limit,
                period=period,
            ),
        )

    if function := function:
        return _wrap(function)

    else:
        return _wrap


class _AsyncThrottle[**Args, Result]:
    __slots__ = (
        "__annotations__",
        "__defaults__",
        "__doc__",
        "__globals__",
        "__kwdefaults__",
        "__name__",
        "__qualname__",
        "__wrapped__",
        "_entries",
        "_function",
        "_limit",
        "_lock",
        "_period",
    )

    def __init__(
        self,
        function: Callable[Args, Coroutine[Any, Any, Result]],
        /,
        limit: int,
        period: timedelta | float,
    ) -> None:
        self._function: Callable[Args, Coroutine[Any, Any, Result]] = function
        self._entries: deque[float] = deque()
        self._lock: Lock = Lock()
        self._limit: int = limit
        self._period: float
        match period:
            case timedelta() as delta:
                self._period = delta.total_seconds()

            case period_seconds:
                self._period = period_seconds

        # mimic function attributes if able
        mimic_function(function, within=self)

    async def __call__(
        self,
        *args: Args.args,
        **kwargs: Args.kwargs,
    ) -> Result:
        async with self._lock:
            time_now: float = monotonic()
            while self._entries:  # cleanup old entries
                if self._entries[0] + self._period <= time_now:
                    self._entries.popleft()

                else:
                    break

            if len(self._entries) >= self._limit:
                await sleep(self._entries[0] - time_now)

            self._entries.append(monotonic())

        return await self._function(*args, **kwargs)



from typing import Final

from haiway import getenv_str

__all__ = [
    "POSTGRES_DATABASE",
    "POSTGRES_HOST",
    "POSTGRES_PASSWORD",
    "POSTGRES_PORT",
    "POSTGRES_SSLMODE",
    "POSTGRES_USER",
]

POSTGRES_DATABASE: Final[str] = getenv_str("POSTGRES_DATABASE", default="postgres")
POSTGRES_HOST: Final[str] = getenv_str("POSTGRES_HOST", default="localhost")
POSTGRES_PORT: Final[str] = getenv_str("POSTGRES_PORT", default="5432")
POSTGRES_USER: Final[str] = getenv_str("POSTGRES_USER", default="postgres")
POSTGRES_PASSWORD: Final[str] = getenv_str("POSTGRES_PASSWORD", default="postgres")
POSTGRES_SSLMODE: Final[str] = getenv_str("POSTGRES_SSLMODE", default="prefer")



from haiway.context.access import ScopeContext, ctx
from haiway.context.disposables import Disposable, Disposables
from haiway.context.identifier import ScopeIdentifier
from haiway.context.metrics import (
    MetricsContext,
    MetricsHandler,
    MetricsReading,
    MetricsRecording,
    MetricsScopeEntering,
    MetricsScopeExiting,
)
from haiway.context.state import StateContext
from haiway.context.types import MissingContext, MissingState

__all__ = [
    "Disposable",
    "Disposables",
    "MetricsContext",
    "MetricsHandler",
    "MetricsReading",
    "MetricsRecording",
    "MetricsScopeEntering",
    "MetricsScopeExiting",
    "MissingContext",
    "MissingState",
    "ScopeContext",
    "ScopeIdentifier",
    "StateContext",
    "ctx",
]



from logging import getLogger

from haiway import MetricsLogger, ctx
from starlette.datastructures import MutableHeaders
from starlette.exceptions import HTTPException, WebSocketException
from starlette.types import ASGIApp, Message, Receive, Scope, Send

__all__ = [
    "ContextMiddleware",
]


class ContextMiddleware:
    def __init__(
        self,
        app: ASGIApp,
    ) -> None:
        self.app = app

    async def __call__(
        self,
        scope: Scope,
        receive: Receive,
        send: Send,
    ) -> None:
        match scope["type"]:
            case "websocket":
                return await self._handle_websocket(scope, receive, send)

            case "http":
                return await self._handle_http(scope, receive, send)

            case _:
                return await self.app(scope, receive, send)

    async def _handle_websocket(
        self,
        scope: Scope,
        receive: Receive,
        send: Send,
    ) -> None:
        async with ctx.scope(
            f"ws {scope['path']}",
            *scope["app"].extra.get("state", ()),
            logger=getLogger("server"),
            metrics=MetricsLogger.handler(),
        ):
            try:
                return await self.app(
                    scope,
                    receive,
                    send,
                )

            except WebSocketException as exc:
                raise exc

            except BaseException as exc:
                error_type: type[BaseException] = type(exc)
                error_message: str = (
                    f"{error_type.__name__} [{error_type.__module__}] - that is an error!"
                )

                if __debug__:
                    import traceback

                    error_message = error_message + f"\n{traceback.format_exc()}"

                ctx.log_error(error_message)

                raise WebSocketException(
                    code=1011,
                    reason=error_message,
                ) from exc

    async def _handle_http(
        self,
        scope: Scope,
        receive: Receive,
        send: Send,
    ) -> None:
        async with ctx.scope(
            f"{scope.get('method', '')} {scope['path']}",
            *scope["app"].extra.get("state", ()),
            logger=getLogger("server"),
            metrics=MetricsLogger.handler(),
        ) as trace_id:

            async def traced_send(message: Message) -> None:
                match message["type"]:
                    case "http.response.start":
                        headers = MutableHeaders(scope=message)
                        headers["trace_id"] = f"{trace_id}"

                    case _:
                        pass

                await send(message)

            try:
                return await self.app(
                    scope,
                    receive,
                    traced_send,
                )

            except HTTPException as exc:
                if exc.headers is None:
                    exc.headers = {"trace_id": f"{trace_id}"}

                elif isinstance(exc.headers, dict):
                    exc.headers["trace_id"] = f"{trace_id}"

                raise exc  # do not change behavior for HTTPException

            except BaseException as exc:
                error_type: type[BaseException] = type(exc)
                error_message: str = (
                    f"{error_type.__name__} [{error_type.__module__}] - that is an error!"
                )

                if __debug__:
                    import traceback

                    error_message = error_message + f"\n{traceback.format_exc()}"

                ctx.log_error(error_message)

                raise HTTPException(
                    status_code=500,
                    headers={"trace_id": f"{trace_id}"},
                    detail=error_message,
                ) from exc



from collections.abc import Mapping, Sequence
from datetime import date, datetime, time
from types import TracebackType
from typing import TYPE_CHECKING, Any, Protocol, runtime_checkable
from uuid import UUID

if TYPE_CHECKING:
    from integrations.postgres.state import PostgresConnection

__all__ = [
    "PostgresConnectionContext",
    "PostgresConnectionPreparing",
    "PostgresException",
    "PostgresStatementExecuting",
    "PostgresTransactionContext",
    "PostgresTransactionPreparing",
    "PostgresValue",
]

type PostgresValue = UUID | datetime | date | time | str | bytes | float | int | bool | None
type PostgresRow = Mapping[str, Any]


@runtime_checkable
class PostgresStatementExecuting(Protocol):
    async def __call__(
        self,
        statement: str,
        /,
        *args: PostgresValue,
    ) -> Sequence[PostgresRow]: ...


@runtime_checkable
class PostgresTransactionContext(Protocol):
    async def __aenter__(self) -> None: ...

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> bool | None: ...


@runtime_checkable
class PostgresTransactionPreparing(Protocol):
    def __call__(self) -> PostgresTransactionContext: ...


@runtime_checkable
class PostgresConnectionContext(Protocol):
    async def __aenter__(self) -> "PostgresConnection": ...

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> bool | None: ...


@runtime_checkable
class PostgresConnectionPreparing(Protocol):
    def __call__(self) -> PostgresConnectionContext: ...


class PostgresException(Exception):
    pass



import types
import typing
from collections.abc import Callable, Mapping, MutableMapping, Sequence
from types import GenericAlias, NoneType, UnionType
from typing import (
    Any,
    ClassVar,
    ForwardRef,
    Generic,
    Literal,
    ParamSpec,
    Self,
    TypeAliasType,
    TypeVar,
    TypeVarTuple,
    _GenericAlias,  # pyright: ignore
    final,
    get_args,
    get_origin,
    get_type_hints,
    is_typeddict,
    overload,
)

from haiway import types as haiway_types
from haiway.types import MISSING, Missing

__all__ = [
    "AttributeAnnotation",
    "attribute_annotations",
    "resolve_attribute_annotation",
]


@final
class AttributeAnnotation:
    __slots__ = (
        "arguments",
        "extra",
        "origin",
        "required",
    )

    def __init__(
        self,
        *,
        origin: Any,
        arguments: Sequence[Any] | None = None,
        required: bool = True,
        extra: Mapping[str, Any] | None = None,
    ) -> None:
        self.origin: Any = origin
        self.arguments: Sequence[Any]
        if arguments is None:
            self.arguments = ()

        else:
            self.arguments = arguments

        self.required: bool = required

        self.extra: Mapping[str, Any]
        if extra is None:
            self.extra = {}

        else:
            self.extra = extra

    def update_required(
        self,
        required: bool,
        /,
    ) -> Self:
        object.__setattr__(
            self,
            "required",
            self.required and required,
        )

        return self

    def __str__(self) -> str:
        if alias := self.extra.get("TYPE_ALIAS"):
            return alias

        origin_str: str = getattr(self.origin, "__name__", str(self.origin))
        arguments_str: str
        if self.arguments:
            arguments_str = "[" + ", ".join(str(arg) for arg in self.arguments) + "]"

        else:
            arguments_str = ""

        if module := getattr(self.origin, "__module__", None):
            return f"{module}.{origin_str}{arguments_str}"

        else:
            return f"{origin_str}{arguments_str}"


def attribute_annotations(
    cls: type[Any],
    /,
    type_parameters: Mapping[str, Any],
) -> Mapping[str, AttributeAnnotation]:
    self_annotation = AttributeAnnotation(
        origin=cls,
        # ignore arguments here, State (and draive.DataModel) will have them resolved at this stage
        arguments=[],
    )

    # ignore args_keys here, State (and draive.DataModel) will have them resolved at this stage
    recursion_guard: MutableMapping[str, AttributeAnnotation] = {
        _recursion_key(cls, default=str(self_annotation)): self_annotation
    }

    attributes: dict[str, AttributeAnnotation] = {}
    for key, annotation in get_type_hints(cls, localns={cls.__name__: cls}).items():
        # do not include private or special items
        if key.startswith("_"):
            continue

        # do not include ClassVars
        if (get_origin(annotation) or annotation) is ClassVar:
            continue

        attributes[key] = resolve_attribute_annotation(
            annotation,
            type_parameters=type_parameters,
            module=cls.__module__,
            self_annotation=self_annotation,
            recursion_guard=recursion_guard,
        )

    return attributes


def _resolve_none(
    annotation: Any,
) -> AttributeAnnotation:
    return AttributeAnnotation(origin=NoneType)


def _resolve_missing(
    annotation: Any,
) -> AttributeAnnotation:
    # special case - attributes marked as missing are not required
    # Missing does not work properly within TypedDict though
    return AttributeAnnotation(
        origin=Missing,
        required=False,
    )


def _resolve_literal(
    annotation: Any,
) -> AttributeAnnotation:
    return AttributeAnnotation(
        origin=Literal,
        arguments=get_args(annotation),
    )


def _resolve_forward_ref(
    annotation: ForwardRef | str,
    /,
    module: str,
    type_parameters: Mapping[str, Any],
    self_annotation: AttributeAnnotation | None,
    recursion_guard: MutableMapping[str, AttributeAnnotation],
) -> AttributeAnnotation:
    forward_ref: ForwardRef
    match annotation:
        case str() as string:
            forward_ref = ForwardRef(string, module=module)

        case reference:
            forward_ref = reference

    if evaluated := forward_ref._evaluate(
        globalns=None,
        localns=None,
        recursive_guard=frozenset(),
    ):
        return resolve_attribute_annotation(
            evaluated,
            type_parameters=type_parameters,
            module=module,
            self_annotation=self_annotation,
            recursion_guard=recursion_guard,
        )

    else:
        raise RuntimeError(f"Cannot resolve annotation of {annotation}")


def _resolve_generic_alias(  # noqa: PLR0911, PLR0912
    annotation: GenericAlias,
    /,
    module: str,
    type_parameters: Mapping[str, Any],
    self_annotation: AttributeAnnotation | None,
    recursion_guard: MutableMapping[str, AttributeAnnotation],
) -> AttributeAnnotation:
    match get_origin(annotation):
        case TypeAliasType() as alias:  # pyright: ignore[reportUnnecessaryComparison]
            return _resolve_type_alias(
                alias,
                type_parameters={
                    # verify if we should pass all parameters
                    param.__name__: get_args(annotation)[idx]
                    for idx, param in enumerate(alias.__type_params__)
                },
                module=module,
                self_annotation=self_annotation,
                recursion_guard=recursion_guard,
            )

        case origin if issubclass(origin, Generic):
            match origin.__class_getitem__(  # pyright: ignore[reportUnknownMemberType, reportAttributeAccessIssue]
                tuple(
                    type_parameters.get(
                        arg.__name__,
                        arg.__bound__ or Any,
                    )
                    if isinstance(arg, TypeVar)
                    else arg
                    for arg in get_args(annotation)
                )
            ):
                case GenericAlias() as generic_alias:
                    resolved_attribute = AttributeAnnotation(origin=generic_alias.__origin__)
                    if recursion_key := _recursion_key(generic_alias):
                        if recursive := recursion_guard.get(recursion_key):
                            return recursive

                        else:
                            recursion_guard[recursion_key] = resolved_attribute

                    resolved_attribute.arguments = [
                        resolve_attribute_annotation(
                            argument,
                            type_parameters=type_parameters,
                            module=module,
                            self_annotation=self_annotation,
                            recursion_guard=recursion_guard,
                        )
                        for argument in get_args(generic_alias)
                    ]

                    return resolved_attribute

                # use resolved type if it is not an alias again
                case resolved:  # pyright: ignore
                    resolved_attribute = AttributeAnnotation(origin=resolved)

                    if recursion_key := _recursion_key(origin):
                        if recursive := recursion_guard.get(recursion_key):
                            return recursive

                        else:
                            recursion_guard[recursion_key] = resolved_attribute

                    resolved_attribute.arguments = [
                        resolve_attribute_annotation(
                            argument,
                            type_parameters=type_parameters,
                            module=module,
                            self_annotation=self_annotation,
                            recursion_guard=recursion_guard,
                        )
                        for argument in get_args(annotation)
                    ]

                    return resolved_attribute

        case origin:
            resolved_attribute = AttributeAnnotation(origin=origin)

            if recursion_key := _recursion_key(origin):
                if recursive := recursion_guard.get(recursion_key):
                    return recursive

            resolved_attribute.arguments = [
                resolve_attribute_annotation(
                    argument,
                    type_parameters=type_parameters,
                    module=module,
                    self_annotation=self_annotation,
                    recursion_guard=recursion_guard,
                )
                for argument in get_args(annotation)
            ]

            return resolved_attribute


def _resolve_special_generic_alias(
    annotation: Any,
    /,
    module: str,
    type_parameters: Mapping[str, Any],
    self_annotation: AttributeAnnotation | None,
    recursion_guard: MutableMapping[str, AttributeAnnotation],
) -> AttributeAnnotation:
    origin: type[Any] = get_origin(annotation)
    resolved_attribute = AttributeAnnotation(origin=origin)

    if recursion_key := _recursion_key(origin):
        if recursive := recursion_guard.get(recursion_key):
            return recursive

        else:
            recursion_guard[recursion_key] = resolved_attribute

    resolved_attribute.arguments = [
        resolve_attribute_annotation(
            argument,
            type_parameters=type_parameters,
            module=module,
            self_annotation=self_annotation,
            recursion_guard=recursion_guard,
        )
        for argument in get_args(annotation)
    ]

    return resolved_attribute


def _resolve_type_alias(
    annotation: TypeAliasType,
    /,
    module: str,
    type_parameters: Mapping[str, Any],
    self_annotation: AttributeAnnotation | None,
    recursion_guard: MutableMapping[str, AttributeAnnotation],
) -> AttributeAnnotation:
    resolved_attribute = AttributeAnnotation(origin=MISSING)

    if recursion_key := _recursion_key(annotation):
        if recursive := recursion_guard.get(recursion_key):
            return recursive

        else:
            recursion_guard[recursion_key] = resolved_attribute

    resolved: AttributeAnnotation = resolve_attribute_annotation(
        annotation.__value__,
        module=annotation.__module__ or module,
        type_parameters=type_parameters,
        self_annotation=self_annotation,
        recursion_guard=recursion_guard,
    )

    resolved_attribute.origin = resolved.origin
    resolved_attribute.arguments = resolved.arguments
    resolved_attribute.extra = {
        **resolved.extra,
        "TYPE_ALIAS": annotation.__name__,
    }
    resolved_attribute.required = resolved.required

    return resolved_attribute


def _resolve_type_var(
    annotation: TypeVar,
    /,
    module: str,
    type_parameters: Mapping[str, Any],
    self_annotation: AttributeAnnotation | None,
    recursion_guard: MutableMapping[str, AttributeAnnotation],
) -> AttributeAnnotation:
    return resolve_attribute_annotation(
        type_parameters.get(
            annotation.__name__,
            # use bound as default or Any otherwise
            annotation.__bound__ or Any,
        ),
        module=module,
        type_parameters=type_parameters,
        self_annotation=self_annotation,
        recursion_guard=recursion_guard,
    )


def _resolve_type_union(
    annotation: UnionType,
    /,
    module: str,
    type_parameters: Mapping[str, Any],
    self_annotation: AttributeAnnotation | None,
    recursion_guard: MutableMapping[str, AttributeAnnotation],
) -> AttributeAnnotation:
    arguments: Sequence[AttributeAnnotation] = [
        resolve_attribute_annotation(
            argument,
            type_parameters=type_parameters,
            module=module,
            self_annotation=self_annotation,
            recursion_guard=recursion_guard,
        )
        for argument in get_args(annotation)
    ]
    return AttributeAnnotation(
        origin=UnionType,  # pyright: ignore[reportArgumentType]
        arguments=arguments,
        required=all(argument.required for argument in arguments),
    )


def _resolve_callable(
    annotation: Any,
    /,
    module: str,
    type_parameters: Mapping[str, Any],
    self_annotation: AttributeAnnotation | None,
    recursion_guard: MutableMapping[str, AttributeAnnotation],
) -> AttributeAnnotation:
    return AttributeAnnotation(
        origin=Callable,
        arguments=[
            resolve_attribute_annotation(
                argument,
                type_parameters=type_parameters,
                module=module,
                self_annotation=self_annotation,
                recursion_guard=recursion_guard,
            )
            for argument in get_args(annotation)
        ],
    )


def _resolve_type_box(
    annotation: Any,
    /,
    module: str,
    type_parameters: Mapping[str, Any],
    self_annotation: AttributeAnnotation | None,
    recursion_guard: MutableMapping[str, AttributeAnnotation],
) -> AttributeAnnotation:
    return resolve_attribute_annotation(
        get_args(annotation)[0],
        type_parameters=type_parameters,
        module=module,
        self_annotation=self_annotation,
        recursion_guard=recursion_guard,
    )


def _resolve_type_not_required(
    annotation: Any,
    /,
    module: str,
    type_parameters: Mapping[str, Any],
    self_annotation: AttributeAnnotation | None,
    recursion_guard: MutableMapping[str, AttributeAnnotation],
) -> AttributeAnnotation:
    return resolve_attribute_annotation(
        get_args(annotation)[0],
        type_parameters=type_parameters,
        module=module,
        self_annotation=self_annotation,
        recursion_guard=recursion_guard,
    ).update_required(False)


def _resolve_type_optional(
    annotation: Any,
    /,
    module: str,
    type_parameters: Mapping[str, Any],
    self_annotation: AttributeAnnotation | None,
    recursion_guard: MutableMapping[str, AttributeAnnotation],
) -> AttributeAnnotation:
    return AttributeAnnotation(
        origin=UnionType,  # pyright: ignore[reportArgumentType]
        arguments=[
            resolve_attribute_annotation(
                get_args(annotation)[0],
                type_parameters=type_parameters,
                module=module,
                self_annotation=self_annotation,
                recursion_guard=recursion_guard,
            ),
            AttributeAnnotation(origin=NoneType),
        ],
    )


def _resolve_type_typeddict(
    annotation: Any,
    /,
    module: str,
    type_parameters: Mapping[str, Any],
    self_annotation: AttributeAnnotation | None,
    recursion_guard: MutableMapping[str, AttributeAnnotation],
) -> AttributeAnnotation:
    resolved_attribute = AttributeAnnotation(origin=annotation)

    if recursion_key := _recursion_key(annotation):
        if recursive := recursion_guard.get(recursion_key):
            return recursive

        else:
            recursion_guard[recursion_key] = resolved_attribute

    resolved_attribute.arguments = [
        resolve_attribute_annotation(
            argument,
            type_parameters=type_parameters,
            module=module,
            self_annotation=self_annotation,
            recursion_guard=recursion_guard,
        )
        for argument in get_args(annotation)
    ]

    attributes: dict[str, AttributeAnnotation] = {}
    for key, element in get_type_hints(
        annotation,
        localns={annotation.__name__: annotation},
    ).items():
        attributes[key] = resolve_attribute_annotation(
            element,
            type_parameters=type_parameters,
            module=getattr(annotation, "__module__", module),
            self_annotation=resolved_attribute,
            recursion_guard=recursion_guard,
        ).update_required(key in annotation.__required_keys__)
    resolved_attribute.extra = {
        "attributes": attributes,
        "required": annotation.__required_keys__,
    }
    return resolved_attribute


def _resolve_type(
    annotation: Any,
    /,
    module: str,
    type_parameters: Mapping[str, Any],
    self_annotation: AttributeAnnotation | None,
    recursion_guard: MutableMapping[str, AttributeAnnotation],
) -> AttributeAnnotation:
    if recursion_key := _recursion_key(annotation):
        if recursive := recursion_guard.get(recursion_key):
            return recursive

        # not updating recursion guard here - it might be a builtin type

    return AttributeAnnotation(
        origin=annotation,
        arguments=[
            resolve_attribute_annotation(
                argument,
                type_parameters=type_parameters,
                module=module,
                self_annotation=self_annotation,
                recursion_guard=recursion_guard,
            )
            for argument in get_args(annotation)
        ],
    )


def resolve_attribute_annotation(  # noqa: C901, PLR0911, PLR0912
    annotation: Any,
    /,
    module: str,
    type_parameters: Mapping[str, Any],
    self_annotation: AttributeAnnotation | None,
    recursion_guard: MutableMapping[str, AttributeAnnotation],
) -> AttributeAnnotation:
    match get_origin(annotation) or annotation:
        case types.NoneType | None:
            return _resolve_none(
                annotation=annotation,
            )

        case haiway_types.Missing:
            return _resolve_missing(
                annotation=annotation,
            )

        case types.UnionType | typing.Union:
            return _resolve_type_union(
                annotation,
                module=module,
                type_parameters=type_parameters,
                self_annotation=self_annotation,
                recursion_guard=recursion_guard,
            )

        case typing.Literal:
            return _resolve_literal(annotation)

        case typeddict if is_typeddict(typeddict):
            return _resolve_type_typeddict(
                typeddict,
                module=module,
                type_parameters=type_parameters,
                self_annotation=self_annotation,
                recursion_guard=recursion_guard,
            )

        case typing.Callable:  # pyright: ignore
            return _resolve_callable(
                annotation,
                module=module,
                type_parameters=type_parameters,
                self_annotation=self_annotation,
                recursion_guard=recursion_guard,
            )

        case typing.Annotated | typing.Final | typing.Required:
            return _resolve_type_box(
                annotation,
                module=module,
                type_parameters=type_parameters,
                self_annotation=self_annotation,
                recursion_guard=recursion_guard,
            )

        case typing.NotRequired:
            return _resolve_type_not_required(
                annotation,
                module=module,
                type_parameters=type_parameters,
                self_annotation=self_annotation,
                recursion_guard=recursion_guard,
            )

        case typing.Optional:  # optional is a Union[Value, None]
            return _resolve_type_optional(
                annotation,
                module=module,
                type_parameters=type_parameters,
                self_annotation=self_annotation,
                recursion_guard=recursion_guard,
            )

        case typing.Self:  # pyright: ignore
            if self_annotation:
                return self_annotation

            else:
                raise RuntimeError(f"Unresolved Self annotation: {annotation}")

        case _:
            match annotation:
                case str() | ForwardRef():
                    return _resolve_forward_ref(
                        annotation,
                        module=module,
                        type_parameters=type_parameters,
                        self_annotation=self_annotation,
                        recursion_guard=recursion_guard,
                    )

                case GenericAlias():
                    return _resolve_generic_alias(
                        annotation,
                        module=module,
                        type_parameters=type_parameters,
                        self_annotation=self_annotation,
                        recursion_guard=recursion_guard,
                    )

                case _GenericAlias():
                    return _resolve_special_generic_alias(
                        annotation,
                        module=module,
                        type_parameters=type_parameters,
                        self_annotation=self_annotation,
                        recursion_guard=recursion_guard,
                    )

                case TypeAliasType():
                    return _resolve_type_alias(
                        annotation,
                        module=module,
                        type_parameters=type_parameters,
                        self_annotation=self_annotation,
                        recursion_guard=recursion_guard,
                    )

                case TypeVar():
                    return _resolve_type_var(
                        annotation,
                        module=module,
                        type_parameters=type_parameters,
                        self_annotation=self_annotation,
                        recursion_guard=recursion_guard,
                    )

                case ParamSpec():
                    raise NotImplementedError(f"Unresolved ParamSpec annotation: {annotation}")

                case TypeVarTuple():
                    raise NotImplementedError(f"Unresolved TypeVarTuple annotation: {annotation}")

                case _:  # finally use whatever there was
                    return _resolve_type(
                        annotation,
                        module=module,
                        type_parameters=type_parameters,
                        self_annotation=self_annotation,
                        recursion_guard=recursion_guard,
                    )


@overload
def _recursion_key(
    annotation: Any,
    /,
) -> str | None: ...


@overload
def _recursion_key(
    annotation: Any,
    /,
    default: str,
) -> str: ...


def _recursion_key(
    annotation: Any,
    /,
    default: str | None = None,
) -> str | None:
    args_suffix: str
    if arguments := get_args(annotation):
        arguments_string: str = ", ".join(
            _recursion_key(
                argument,
                default="?",
            )
            for argument in arguments
        )
        args_suffix = f"[{arguments_string}]"

    else:
        args_suffix = ""

    if qualname := getattr(annotation, "__qualname__", None):
        return qualname + args_suffix

    module_prefix: str
    if module := getattr(annotation, "__module__", None):
        module_prefix = module + "."

    else:
        module_prefix = ""

    if name := getattr(annotation, "__name__", None):
        return module_prefix + name + args_suffix

    return default
